class TestWsgi:
    def test_module_loads_config(self, flask_app_test, config):
        assert config.initialized is False
        # needed to avoid config postprocessing side effects in this test
        with flask_app_test.test_request_context():
            from palladium import wsgi  # initializes config

        # fit mode should only be set by non-server scripts
        assert config.get('__mode__') is None
        assert config.initialized is True



"""Module to be used for exposing the web service with other WSGI
servers. It initializes Palladium's configuration and provides the
Flask object.
"""
from .util import initialize_config
from .server import app

# Initialization is needed to obtain same behavior of pld-devserver
# and other WSGI servers (e.g., gunicorn). Without this initialization
# call, initialization would be delayed until the first request arrives
initialize_config()



""":class:`~palladium.interfaces.ModelPersister` implementations.
"""
import logging
from abc import abstractmethod
import base64
from contextlib import contextmanager
import gzip
import io
import json
import os
import pickle
import codecs
from pkg_resources import parse_version
from tempfile import TemporaryFile
from threading import Lock

import requests
from sqlalchemy import create_engine
from sqlalchemy import CLOB
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import LargeBinary
from sqlalchemy import String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool
from sqlalchemy.types import TypeDecorator

from . import __version__
from .interfaces import annotate
from .interfaces import ModelPersister
from .util import logger
from .util import PluggableDecorator
from .util import process_store
from .util import RruleThread
from .util import session_scope


class UpgradeSteps:
    def __init__(self):
        self.steps = []

    def add(self, version):
        def decorator(func):
            self.steps.append((parse_version(version), func))
            return func
        return decorator

    def run(self, persister, from_version, to_version):
        from_version = parse_version(from_version)
        to_version = parse_version(to_version)
        results = []
        for version, func in sorted(self.steps):
            if from_version < version <= to_version:
                results.append(func(persister))
        return results


class FileLikeIO:
    """Used by :class:`FileLike` to access low level file handle
    operations.
    """

    @abstractmethod
    def open(self, path, mode='r'):
        """Return a file handle

        For normal files, the implementation is:

        ```python
        return open(path, mode)
        ```
        """

    @abstractmethod
    def exists(self, path):
        """Test whether a path exists

        For normal files, the implementation is:

        ```python
        return os.path.exists(path)
        ```
        """

    @abstractmethod
    def remove(self, path):
        """Remove a file

        For normal files, the implementation is:

        ```python
        os.remove(path)
        ```
        """


class FileIO(FileLikeIO):
    def open(self, path, mode='r'):
        return open(path, mode)

    def exists(self, path):
        return os.path.exists(path)

    def remove(self, path):
        os.remove(path)


class RestIO(FileLikeIO):
    def __init__(self, auth):
        self.session = requests.Session()
        self.session.auth = auth

    @contextmanager
    def _write(self, url, mode):
        # Use a context manager to send the actual request out once
        # the file that FileLike writes into is 'closed'.
        if '+' not in mode:
            mode += '+'
        with TemporaryFile(mode=mode) as fh:
            yield fh
            fh.seek(0)
            res = self.session.put(url, data=fh)
        res.raise_for_status()

    def open(self, path, mode='r'):
        if mode[0] == 'r':
            res = self.session.get(path, stream=True)
            res.raise_for_status()
            if 'b' in mode:
                return res.raw
            else:
                reader = codecs.getreader(res.encoding or 'utf-8')
                return reader(res.raw)
        elif mode == 'wb':
            return self._write(path, mode=mode)
        raise NotImplementedError("filemode: %s" % (mode,))

    def exists(self, path):
        res = self.session.head(path)
        if res.status_code == 404:
            return False
        res.raise_for_status()
        return True

    def remove(self, path):
        res = self.session.delete(path)
        res.raise_for_status()


class FileLike(ModelPersister):
    """A :class:`~palladium.interfaces.ModelPersister` that pickles
    models through file-like handles.

    An argument ``io`` is used to access low level file handle
    operations.
    """
    upgrade_steps = UpgradeSteps()

    def __init__(self, path, io):
        """
        :param str path:
          The *path* template that I will use to store models,
          e.g. ``/path/to/model-{version}``.

        :param FileLikeIO io:
          Used to access low level file handle operations.
        """
        if '{version}' not in path:
            raise ValueError(
                "Your file persister path must have a {version} placeholder,"
                "e.g., model-{version}.pickle."
                )
        self.path = path
        self.io = io

    def read(self, version=None):
        use_active_model = version is None

        if version is None:
            props = self.list_properties()
            if 'active-model' not in props:
                raise LookupError("No active model available: " + self.path)
            version = props['active-model']

        fname = self.path.format(version=version) + '.pkl.gz'
        if not self.io.exists(fname):
            if use_active_model:
                raise LookupError(
                    "Activated model not available. Maybe it was deleted.")
            else:
                raise LookupError("No such version: {}".format(version))

        with self.io.open(fname, 'rb') as fh:
            with gzip.open(fh, 'rb') as f:
                model = pickle.load(f)

        attachments = annotate(model).get('__attachments__', [])
        for key in attachments:
            fname_attach = self.attach_filename(version=version, key=key)
            if self.io.exists(fname_attach):
                with open(fname_attach, 'rb') as f:
                    data_attach = base64.b64encode(f.read())
                    annotate(model, {key: data_attach})

        return model

    def write(self, model):
        last_version = 0
        li = self.list_models()
        if li:
            last_version = li[-1]['version']

        version = last_version + 1
        li.append(annotate(model, {'version': version}))

        annotations = annotate(model)
        attachments = {
            key: data
            for key, data in annotations.items()
            if key.startswith('attachments/')
            }
        if attachments:
            for key in attachments:
                del annotations[key]
            annotations['__attachments__'] = tuple(attachments.keys())
        annotate(model, annotations)

        fname = self.path.format(version=version) + '.pkl.gz'
        with self.io.open(fname, 'wb') as fh:
            with gzip.open(fh, 'wb') as f:
                pickle.dump(model, f)

        if attachments:
            for key, data in attachments.items():
                fname_attach = self.attach_filename(version=version, key=key)
                with self.io.open(fname_attach, 'wb') as f:
                    f.write(base64.b64decode(data))

        self._update_md({'models': li})
        return version

    def list_models(self):
        return self._read_md()['models']

    def list_properties(self):
        return self._read_md()['properties']

    def activate(self, version):
        md = self._read_md()
        md['properties']['active-model'] = str(version)
        versions = [m['version'] for m in md['models']]
        if int(version) not in versions:
            raise LookupError("No such version: {}".format(version))
        self._update_md({'properties': md['properties']})

    def delete(self, version):
        version = int(version)
        md = self._read_md()
        try:
            model_md = [m for m in md['models'] if m['version'] == version][0]
        except IndexError:
            raise LookupError("No such version: {}".format(version))

        self._update_md({
            'models': [m for m in md['models'] if m['version'] != version]})
        self.io.remove(self.path.format(version=version) + '.pkl.gz')

        attachments = model_md.get('__attachments__', [])
        for key in attachments:
            fname_attach = self.attach_filename(version=version, key=key)
            if self.io.exists(fname_attach):
                self.io.remove(fname_attach)

    @property
    def _md_filename(self):
        return self.path.format(version='metadata') + '.json'

    def attach_filename(self, version, key):
        return (
            self.path.format(version=version) +
            '-{}'.format(key[len('attachments/'):])
            )

    def _read_md(self):
        if self.io.exists(self._md_filename):
            with self.io.open(self._md_filename, 'r') as f:
                return json.load(f)
        return {'models': [], 'properties': {'db-version': __version__}}

    def _update_md(self, data):
        data2 = self._read_md()
        data2.update(data)
        with self.io.open(self._md_filename, 'wb') as f:
            bytes = json.dumps(data2, indent=4).encode('utf-8')
            f.write(bytes)

    def upgrade(self, from_version=None, to_version=__version__):
        if from_version is None:
            try:
                from_version = self._read_md()['properties']['db-version']
            except (KeyError, TypeError):
                from_version = "0.0"

        self.upgrade_steps.run(self, from_version, to_version)
        md = self._read_md()
        md['properties']['db-version'] = to_version
        self._update_md(md)

    @upgrade_steps.add('1.0')
    def _upgrade_1_0(self):
        if self.io.exists(self._md_filename):
            with self.io.open(self._md_filename, 'r') as f:
                old_md = json.load(f)
        else:
            old_md = None

        active_model = old_md[-1]['version'] if old_md else None
        new_md = {
            'models': old_md or [],
            'properties': {},
            }
        if active_model is not None:
            new_md['properties']['active-model'] = str(active_model)

        with self.io.open(self._md_filename, 'w') as f:
            json.dump(new_md, f, indent=4)


class File(FileLike):
    """A :class:`~palladium.interfaces.ModelPersister` that pickles models
    onto the file system, into a given directory.
    """
    def __init__(self, path):
        """
        :param str path:
          The *path* template that I will use to store models,
          e.g. ``/path/to/model-{version}``.
        """
        super().__init__(path, FileIO())


class Rest(FileLike):
    def __init__(self, url, auth):
        super().__init__(url, RestIO(auth))


class Database(ModelPersister):
    """A :class:`~palladium.interfaces.ModelPersister` that pickles models
    into an SQL database.
    """

    upgrade_steps = UpgradeSteps()

    def __init__(
            self, url, poolclass=None, chunk_size=1024 ** 2 * 100,
            table_postfix=''):
        """
        :param str url:
          The database *url* that'll be used to make a connection.
          Format follows RFC-1738.  I'll create a table ``models`` to
          store the pickles in if it doesn't exist yet.

        :param sqlalchemy.pool.Pool poolclass:
          A class specifying DB connection behavior of the engine. If set to
          None, the NullPool will be used.

        :param int chunk_size:
          The pickled contents of the model are stored inside the
          database in chunks.  The default size is 1024 ** 2 * 100
          (100MB).

        :param str table_postfix:
          If *table_postfix* is provided, I will append it to the
          table name of all tables used in this instance.
        """
        if not poolclass:
            poolclass = NullPool
        engine = create_engine(url, poolclass=poolclass)
        self.engine = engine
        self.chunk_size = chunk_size
        self.table_postfix = table_postfix
        self.write_lock = Lock()
        orms = self.create_orm_classes()
        self.Property = orms['Property']
        self.DBModel = orms['DBModel']
        self.DBModelChunk = orms['DBModelChunk']
        metadata = self.DBModel.metadata
        metadata.bind = engine
        metadata.create_all()
        self.session = scoped_session(sessionmaker(bind=engine))
        self._initialize_properties()

    def _initialize_properties(self):
        with session_scope(self.session) as session:
            if session.query(self.Property).count() == 0:
                self._set_property('db-version', __version__)

    def _table_postfix(self, name):
        if self.table_postfix:
            return '{}_{}'.format(name, self.table_postfix)
        else:
            return name

    def create_orm_classes(self):
        Base = declarative_base()

        return {
            'Base': Base,
            'Property': self.PropertyClass(Base),
            'DBModel': self.DBModelClass(Base),
            'DBModelChunk': self.DBModelChunkClass(Base),
            }

    def PropertyClass(self, Base):
        class Property(Base):
            __tablename__ = self._table_postfix('properties')
            id = Column(Integer, primary_key=True)
            name = Column(String(length=10 ** 3))
            value = Column(String(length=10 ** 3), nullable=False)
        return Property

    def DBModelClass(self, Base):
        class DBModel(Base):
            __tablename__ = self._table_postfix('models')
            version = Column(Integer, primary_key=True)
            metadata_ = Column(
                'metadata', String(length=10 ** 6), nullable=False)
            chunks = relationship(
                'DBModelChunk',
                order_by="DBModelChunk.id",
                )
        return DBModel

    def DBModelChunkClass(self, Base):
        class DBModelChunk(Base):
            __tablename__ = self._table_postfix('model_chunks')
            id = Column(Integer, primary_key=True)
            model_version = Column(
                ForeignKey('{}.version'.format(self._table_postfix('models'))))
            blob = Column(LargeBinary, nullable=False)
        return DBModelChunk

    def read(self, version=None):
        use_active_model = version is None

        with session_scope(self.session) as session:
            query = session.query(self.DBModel)
            if not version:
                version = self._active_version
            dbmodel = query.filter_by(version=version).first()

            if dbmodel is not None:
                query2 = session.query(self.DBModelChunk).filter_by(
                    model_version=dbmodel.version
                    ).order_by('id').yield_per(4)
                fileobj = io.BytesIO()
                for chunk in query2:
                    fileobj.write(chunk.blob)
                fileobj.seek(0)
                return pickle.load(gzip.GzipFile(fileobj=fileobj, mode='rb'))

        if use_active_model and dbmodel is None and version is not None:
            raise LookupError(
                "Activated model not available. Maybe it was deleted.")

        raise LookupError("No model available")

    def write(self, model):
        with self.write_lock:
            return self._write(model)

    def _write(self, model):
        max_version = self._get_max_version()
        if not max_version:
            max_version = 0
        version = max_version + 1

        annotate(model, {'version': version})

        fileobj = io.BytesIO()
        pickle.dump(model, gzip.GzipFile(fileobj=fileobj, mode='wb'))
        data = fileobj.getbuffer()
        chunks = [data[i:i + self.chunk_size]
                  for i in range(0, len(data), self.chunk_size)]

        dbmodel = self.DBModel(
            version=version,
            chunks=[self.DBModelChunk(blob=chunk) for chunk in chunks],
            metadata_=json.dumps(model.__metadata__),
            )

        with session_scope(self.session) as session:
            session.add(dbmodel)

        return version

    def _get_max_version(self):
        # We retrieve the max version by hand instead of using an
        # auto-increment because we want to annotate the version
        # number onto the model's metadata.
        with session_scope(self.session) as session:
            query = session.query(self.DBModel.version)
            result = query.order_by(self.DBModel.version.desc()).first()

        if result is not None:
            return result[0]

    def list_models(self):
        with session_scope(self.session) as session:
            results = session.query(self.DBModel.metadata_).all()
        infos = [json.loads(res[0]) for res in results]
        return sorted(infos, key=lambda x: x['version'])

    def list_properties(self):
        with session_scope(self.session) as session:
            return {prop.name: prop.value
                    for prop in session.query(self.Property)}

    def activate(self, version):
        self._set_property('active-model', str(version))

    def delete(self, version):
        with session_scope(self.session) as session:
            session.query(self.DBModel).filter_by(version=version).delete()

    @property
    def _active_version(self):
        with session_scope(self.session) as session:
            active_model = session.query(self.Property).filter_by(
                name='active-model').first()
            if active_model is not None:
                return int(active_model.value)

    def _set_property(self, name, value):
        with session_scope(self.session) as session:
            prop = session.query(self.Property).filter_by(name=name).first()
            if prop is None:
                session.add(self.Property(name=name, value=str(value)))
            else:
                prop.value = str(value)
                session.add(prop)

    def upgrade(self, from_version=None, to_version=__version__):
        if from_version is None:
            from_version = self.list_properties().get('db-version', '0.0')

        self.upgrade_steps.run(self, from_version, to_version)
        self._set_property('db-version', to_version)

    @upgrade_steps.add('1.0')
    def _upgrade_1_0(self):
        if self.list_properties().get('active-model') is None:
            models = self.list_models()
            if models:
                self.activate(int(models[-1]['version']))


class DatabaseCLOB(Database):
    """A :class:`~palladium.interfaces.ModelPersister` derived from
    :class:`Database`, with only the slight difference of using
    CLOB instead of BLOB to store the pickle data.

    Use when BLOB is not available.
    """
    class BytesToBase64Type(TypeDecorator):
        impl = CLOB

        def process_bind_param(self, value, dialect):
            if value is not None:
                value = base64.b64encode(bytes(value)).decode('ascii')
            return value

        def process_result_value(self, value, dialect):
            if value is not None:
                value = base64.b64decode(value.encode('ascii'))
            return value

    def DBModelChunkClass(self, Base):
        class DBModelChunk(Base):
            __tablename__ = self._table_postfix('model_chunks')
            id = Column(Integer, primary_key=True)
            model_version = Column(
                ForeignKey('{}.version'.format(self._table_postfix('models'))))
            blob = Column(self.BytesToBase64Type(String()), nullable=False)
        return DBModelChunk


class CachedUpdatePersister(ModelPersister):
    """A :class:`~palladium.interfaces.ModelPersister` that serves as a
    caching decorator around another `~palladium.interfaces.ModelPersister`
    object.

    Calls to :meth:`~CachedUpdatePersister.read` will look up a model from
    the global ``process_store``, i.e. there is never any actual
    loading involved when calling :meth:`~CachedUpdatePersister.read`.

    To fill the ``process_store`` cache periodically using the return
    value of the underlying :class:`~palladium.interfaces.ModelPersister`'s
    ``read`` method, a dictionary containing keyword arguments to
    :class:`dateutil.rrule.rrule` may be passed.  The cache will then
    be filled periodically according to that recurrence rule.

    If no *update_cache_rrule* is used, the :class:`CachedUpdatePersister`
    will call once and remember the return value of the underlying
    :class:`~palladium.interfaces.ModelPersister`'s ``read`` method during
    initialization.
    """

    cache = process_store
    __pld_config_key__ = 'cachedupdatepersister_default'
    _loaded_version = None

    def __init__(self,
                 impl,
                 update_cache_rrule=None,
                 check_version=True,
                 ):
        """
        :param ModelPersister impl:
          The underlying (decorated) persister object.

        :param dict update_cache_rrule:
          Optional keyword arguments for a
          :class:`dateutil.rrule.rrule` that determines when the cache
          will be updated.  See :class:`~palladium.util.RruleThread` for
          details.

        :param bool check_version:
          If set to `True`, I will perform a check and only load a new
          model from the storage if my cached version differs from
          what's the current active version.
        """
        self.impl = impl
        self.update_cache_rrule = update_cache_rrule
        self.check_version = check_version

    def initialize_component(self, config):
        self.use_cache = config.get('__mode__') != 'fit'
        self.thread = None

        if not self.use_cache:
            return

        self.update_cache()
        logger.info("{}: initial fill of cache done.".format(
            self.__class__.__name__))

        if self.update_cache_rrule:
            self.thread = RruleThread(
                func=self.update_cache, rrule=self.update_cache_rrule)
            self.thread.start()

    def read(self, *args, **kwargs):
        if self.use_cache:
            return self.cache[self.__pld_config_key__]
        else:
            return self.impl.read(*args, **kwargs)

    @PluggableDecorator('update_model_decorators')
    def update_cache(self, *args, **kwargs):
        active_version = None

        if self.check_version:
            active_version = self.list_properties().get('active-model')
            if self._loaded_version == (active_version, args, kwargs):
                return

        try:
            model = self.impl.read(*args, **kwargs)
        except LookupError as ex:
            logging.exception("Cannot find model version")
            model = None
        if model is not None:
            self.cache[self.__pld_config_key__] = model

            if self.check_version:
                self._loaded_version = (active_version, args, kwargs)

            return model

    def write(self, model):
        return self.impl.write(model)

    def list_models(self):
        return self.impl.list_models()

    def list_properties(self):
        return self.impl.list_properties()

    def activate(self, version):
        return self.impl.activate(version)

    def delete(self, version):
        return self.impl.delete(version)

    def upgrade(self, from_version=None, to_version=__version__):
        return self.impl.upgrade(from_version, to_version)


class S3IO(FileLikeIO):
    """IO interface that glues palladium to S3 buckets."""
    def __init__(self, **kwargs):
        try:
            import s3fs
        except ImportError:
            raise ImportError('S3IO needs the s3fs module to work correctly.')

        self.fs = s3fs.S3FileSystem(anon=False)

    def open(self, path, mode='r'):
        # this is needed to avoid reading stale metadata JSONs
        self.fs.invalidate_cache()
        return self.fs.open(path, mode=mode)

    def exists(self, path):
        return self.fs.exists(path)

    def remove(self, path):
        return self.fs.rm(path)


class S3(FileLike):
    """Persister that acts like a File persister but uses
    S3 IO in the background. Expects the path of the bucket
    in the path parameter.

    To use it with palladium, just use this class as model
    persister. For example, if you used the ``File`` persister
    before, change your config as follows:

        -   '!': 'palladium.persistence.File',
        -   'path': 'models/mymodel-{version}',
        +   '!': 'palladium.persistence.S3',
        +   'path': 'your-s3-bucket/models/mymodel-{version}',

    Note that the first part of the path denotes the s3 bucket.

    Parameters
    ----------

    path : str
      The path to the bucket and file, e.g. ``'my-bucket/my-folder/my-file'``.
    """
    def __init__(self, path):
        super().__init__(path, S3IO())



from copy import deepcopy
import logging
from logging.config import dictConfig
import os
import sys
import threading


PALLADIUM_CONFIG_ERROR = """
  Maybe you forgot to set the environment variable PALLADIUM_CONFIG
  to point to your Palladium configuration file?  If so, please
  refer to the manual for more details.
"""

DEFAULT_CONFIG_FILE_LOCATIONS = (
    'palladium-config.py',
    os.path.join('etc', 'palladium-config.py'),
    )


class Config(dict):
    """A dictionary that represents the app's configuration.

    Tries to send a more user friendly message in case of KeyError.
    """
    initialized = False

    def __getitem__(self, name):
        try:
            return super(Config, self).__getitem__(name)
        except KeyError:
            raise KeyError(
                "The required key '{}' was not found in your "
                "configuration. {}".format(name, PALLADIUM_CONFIG_ERROR))


_config = Config()


class ComponentHandler:
    key = '!'

    def __init__(self, config):
        self.config = config
        self.components = []

    def __call__(self, name, props):
        from .util import resolve_dotted_name
        specification = props.copy()
        factory_dotted_name = specification.pop(self.key)
        factory = resolve_dotted_name(factory_dotted_name)
        component = factory(**specification)
        try:
            component.__pld_config_key__ = name
        except AttributeError:
            pass
        self.components.append(component)
        return component

    def finish(self):
        for component in self.components:
            if hasattr(component, 'initialize_component'):
                component.initialize_component(self.config)


class ComponentHandler2(ComponentHandler):
    key = '__factory__'


class CopyHandler:
    key = '__copy__'

    def __init__(self, configs):
        self.configs = configs

    @staticmethod
    def _resolve(configs, dotted_path):
        for config in configs[::-1]:
            value = config
            for part in dotted_path.split('.'):
                try:
                    value = value[part]
                except KeyError:
                    break
            else:
                return value
        else:
            raise KeyError(dotted_path)

    def __call__(self, name, props):
        dotted_path = props[self.key]
        try:
            value = self._resolve(self.configs[-1:], dotted_path)
            self_reference = value is props
        except KeyError:
            self_reference = False

        if self_reference:
            value = self._resolve(self.configs[:-1], dotted_path)
        else:
            try:
                value = self._resolve(self.configs, dotted_path)
            except KeyError:
                if '__default__' in props:
                    return props['__default__']
                else:
                    raise

        value = deepcopy(value)
        nonmagicprops = [
            prop for prop in props
            if not (prop.startswith('__') and prop.endswith('__'))
            ]
        if nonmagicprops:
            recursive_copy = self.key in value
            value.update(props)
            if not recursive_copy:
                del value[self.key]
        return value


class PythonHandler:
    key = '__python__'

    def __init__(self, config):
        self.config = config

    def __call__(self, name, props):
        statements = props.pop(self.key)
        exec(
            statements,
            globals(),
            {key: self.config for key in ['C', 'cfg', 'config']},
            )
        return props


def rewrite_handler(key_from, key_to):
    class RewriteHandler:
        key = key_from
        target = key_to

        def __init__(self, config):
            pass

        def __call__(self, name, props):
            props[self.target] = props.pop(self.key)
            return props
    return RewriteHandler


def _handlers_phase0(configs):
    return {
        Handler.key: Handler(configs) for Handler in [
            rewrite_handler('__factory__', '!'),
            CopyHandler,
            ]
        }


def _handlers_phase1(config):
    return {
        Handler.key: Handler(config) for Handler in [
            PythonHandler,
            ]
        }


def _handlers_phase2(config):
    return {
        Handler.key: Handler(config) for Handler in [
            ComponentHandler,
            ]
        }


def _run_config_handlers_recursive(props, handlers):
    if isinstance(props, dict):
        for key, value in tuple(props.items()):
            if isinstance(value, dict):
                _run_config_handlers_recursive(value, handlers)
                for name, handler in handlers.items():
                    if name in value:
                        value = props[key] = handler(key, value)
            elif isinstance(value, (list, tuple)):
                _run_config_handlers_recursive(value, handlers)
    elif isinstance(props, (list, tuple)):
        for i, item in enumerate(props):
            if isinstance(item, dict):
                _run_config_handlers_recursive(item, handlers)
                for name, handler in handlers.items():
                    if name in item:
                        item = props[i] = handler(str(i), item)
            elif isinstance(item, (list, tuple)):
                _run_config_handlers_recursive(item, handlers)


def _run_config_handlers(config, handlers):
    wrapped_config = {'root': config}
    _run_config_handlers_recursive(wrapped_config, handlers)
    for handler in handlers.values():
        if hasattr(handler, 'finish'):
            handler.finish()
    return wrapped_config['root']


def _initialize_logging(config):
    if 'logging' in config:
        dictConfig(config['logging'])
    else:
        logging.basicConfig(level=logging.DEBUG)


def process_config(
    *configs,
    handlers0=_handlers_phase0,
    handlers1=_handlers_phase1,
    handlers2=_handlers_phase2
):
    config_final = {}

    for config in configs:
        config_org = deepcopy(config_final)
        config_final.update(config)
        _run_config_handlers(
            config_final, handlers0([config_org, config]))
        _run_config_handlers(
            config_final, handlers0([config_final, {}]))

    _run_config_handlers(config_final, handlers1(config_final))
    _run_config_handlers(config_final, handlers2(config_final))
    _initialize_logging(config_final)
    return config_final


_get_config_lock = threading.RLock()


def get_config(**extra):
    with _get_config_lock:
        config = _get_config(**extra)
    return config


def _get_config(**extra):
    if not _config.initialized:
        _config.update(extra)
        _config.initialized = True

        fnames = os.environ.get('PALLADIUM_CONFIG')
        if fnames is None:
            for fname in DEFAULT_CONFIG_FILE_LOCATIONS:
                if os.path.exists(fname):  # pragma: no cover
                    fnames = fname
                    print("Using configuration at {}".format(fname))
                    break

        if fnames is not None:
            configs = []
            fnames = [fname.strip() for fname in fnames.split(',')]
            for fname in fnames:
                sys.path.insert(0, os.path.dirname(fname))
                with open(fname) as f:
                    config = eval(f.read(), {
                        'environ': os.environ,
                        'here': os.path.abspath(os.path.dirname(fname)),
                        })
                configs.append(config)
            _config.update(process_config(_config, *configs))
    return _config


def initialize_config(**extra):
    if _config.initialized:
        raise RuntimeError("Configuration was already initialized")
    return get_config(**extra)



"""HTTP API implementation.
"""

import sys

from docopt import docopt
from flask import Flask
from flask import make_response
from flask import request
import numpy as np
import ujson
from werkzeug.exceptions import BadRequest

from . import __version__
from .fit import activate as activate_base
from .fit import fit as fit_base
from .interfaces import PredictError
from .util import args_from_config
from .util import get_config
from .util import get_metadata
from .util import initialize_config
from .util import logger
from .util import memory_usage_psutil
from .util import PluggableDecorator
from .util import process_store
from .util import run_job
from .util import resolve_dotted_name

app = Flask(__name__)


def make_ujson_response(obj, status_code=200):
    """Encodes the given *obj* to json and wraps it in a response.

    :return:
      A Flask response.
    """
    json_encoded = ujson.encode(obj, ensure_ascii=False)
    resp = make_response(json_encoded)
    resp.mimetype = 'application/json'
    resp.content_type = 'application/json; charset=utf-8'
    resp.status_code = status_code
    return resp


class PredictService:
    """A default :class:`palladium.interfaces.PredictService`
    implementation.

    Aims to work out of the box for the most standard use cases.
    Allows overriding of specific parts of its logic by using granular
    methods to compose the work.
    """
    types = {
        'float': float,
        'int': int,
        'str': str,
        'bool': lambda x: x.lower() == 'true',
        }

    def __init__(
        self,
        mapping,
        params=(),
        entry_point='/predict',
        decorator_list_name='predict_decorators',
        predict_proba=False,
        unwrap_sample=False,
        **kwargs
    ):
        """
        :param mapping:
          A list of query parameters and their type that should be
          included in the request.  These will be processed in the
          :meth:`sample_from_data` method to construct a sample
          that can be used for prediction.  An example that expects
          two request parameters called ``pos`` and ``neg`` that are
          both of type str::

            { ...
              'mapping': [('pos', 'str'), ('neg', 'str')]
            ... }

        :param params:
          Similarly to *mapping*, this is a list of name and type of
          parameters that will be passed to the model's
          :meth:`~palladium.interfaces.Model.predict` method as keyword
          arguments.

        :param predict_proba:
          Instead of returning a single class (the default), when
          *predict_proba* is set to true, the result will instead
          contain a list of class probabilities.

        :param unwrap_sample:
          When working with text, scikit-learn and others will
          sometimes expect the input to be a 1d array of strings
          rather than a 2d array.  Setting *unwrap_sample* to true
          will use this representation.
        """
        self.mapping = mapping
        self.params = params
        self.entry_point = entry_point
        self.decorator_list_name = decorator_list_name
        self.predict_proba = predict_proba
        self.unwrap_sample = unwrap_sample
        vars(self).update(kwargs)

    def initialize_component(self, config):
        create_predict_function(
            self.entry_point, self, self.decorator_list_name, config)

    def __call__(self, model, request):
        try:
            return self.do(model, request)
        except Exception as e:
            return self.response_from_exception(e)

    def do(self, model, request):
        if request.method == 'GET':
            single = True
            samples = np.array([self.sample_from_data(model, request.args)])
        else:
            single = False
            samples = []

            for data in request.json:
                samples.append(self.sample_from_data(model, data))
            samples = np.array(samples)

        params = self.params_from_data(model, request.args)
        y_pred = self.predict(model, samples, **params)
        return self.response_from_prediction(y_pred, single=single)

    def sample_from_data(self, model, data):
        """Convert incoming sample *data* into a numpy array.

        :param model:
          The :class:`~Model` instance to use for making predictions.
        :param data:
          A dict-like with the sample's data, typically retrieved from
          ``request.args`` or similar.
        """
        values = []
        for key, type_name in self.mapping:
            value_type = self.types[type_name]
            values.append(value_type(data[key]))
        if self.unwrap_sample:
            assert len(values) == 1
            return np.array(values[0])
        else:
            return np.array(values, dtype=object)

    def params_from_data(self, model, data):
        """Retrieve additional parameters (keyword arguments) for
        ``model.predict`` from request *data*.

        :param model:
          The :class:`~Model` instance to use for making predictions.
        :param data:
          A dict-like with the parameter data, typically retrieved
          from ``request.args`` or similar.
        """
        params = {}
        for key, type_name in self.params:
            value_type = self.types[type_name]
            if key in data:
                params[key] = value_type(data[key])
            elif hasattr(model, key):
                params[key] = getattr(model, key)
        return params

    def predict(self, model, sample, **kwargs):
        if self.predict_proba:
            return model.predict_proba(sample, **kwargs)
        else:
            return model.predict(sample, **kwargs)

    def response_from_prediction(self, y_pred, single=True):
        """Turns a model's prediction in *y_pred* into a JSON
        response.
        """
        result = y_pred.tolist()
        if single:
            result = result[0]
        response = {
            'metadata': get_metadata(),
            'result': result,
            }
        return make_ujson_response(response, status_code=200)

    def response_from_exception(self, exc):
        if isinstance(exc, PredictError):
            return make_ujson_response({
                'metadata': get_metadata(
                    error_code=exc.error_code,
                    error_message=exc.error_message,
                    status="ERROR"
                )
            }, status_code=500)
        elif isinstance(exc, BadRequest):
            return make_ujson_response({
                'metadata': get_metadata(
                    error_code=-1,
                    error_message="BadRequest: {}".format(exc.args),
                    status="ERROR"
                )
            }, status_code=400)
        else:
            logger.exception("Unexpected error")
            return make_ujson_response({
                'metadata': get_metadata(
                    error_code=-1,
                    error_message="{}: {}".format(
                        exc.__class__.__name__, str(exc)),
                    status="ERROR"
                )
            }, status_code=500)


def predict(model_persister, predict_service):
    try:
        model = model_persister.read()
        response = predict_service(model, request)
    except Exception as exc:
        logger.exception("Unexpected error")
        response = make_ujson_response({
            "status": "ERROR",
            "error_code": -1,
            "error_message": "{}: {}".format(exc.__class__.__name__, str(exc)),
        }, status_code=500)

    return response


@app.route('/alive')
@PluggableDecorator('alive_decorators')
@args_from_config
def alive(alive=None):
    if alive is None:
        alive = {}

    mem, mem_vms = memory_usage_psutil()
    info = {
        'memory_usage': mem,  # rss, resident set size
        'memory_usage_vms': mem_vms,  # vms, virtual memory size
        'palladium_version': __version__,
    }

    info['service_metadata'] = get_config().get('service_metadata', {})

    status_code = 200
    for attr in alive.get('process_store_required', ()):
        obj = process_store.get(attr)
        if obj is not None:
            obj_info = {}
            obj_info['updated'] = process_store.mtime[attr].isoformat()
            if hasattr(obj, '__metadata__'):
                obj_info['metadata'] = obj.__metadata__
            info[attr] = obj_info
        else:
            info[attr] = "N/A"
            status_code = 503

    info['process_metadata'] = process_store['process_metadata']

    return make_ujson_response(info, status_code=status_code)


def create_predict_function(
        route, predict_service, decorator_list_name, config):
    """Creates a predict function and registers it to
    the Flask app using the route decorator.

    :param str route:
      Path of the entry point.

    :param palladium.interfaces.PredictService predict_service:
      The predict service to be registered to this entry point.

    :param str decorator_list_name:
      The decorator list to be used for this predict service. It is
      OK if there is no such entry in the active Palladium config.

    :return:
      A predict service function that will be used to process
      predict requests.
    """
    model_persister = config.get('model_persister')

    @app.route(route, methods=['GET', 'POST'], endpoint=route)
    @PluggableDecorator(decorator_list_name)
    def predict_func():
        return predict(model_persister, predict_service)

    return predict_func


def devserver_cmd(argv=sys.argv[1:]):  # pragma: no cover
    """\
Serve the web API for development.

Usage:
  pld-devserver [options]

Options:
  -h --help               Show this screen.

  --host=<host>           The host to use [default: 0.0.0.0].

  --port=<port>           The port to use [default: 5000].

  --debug=<debug>         Whether or not to use debug mode [default: 0].
"""
    arguments = docopt(devserver_cmd.__doc__, argv=argv)
    initialize_config()
    app.run(
        host=arguments['--host'],
        port=int(arguments['--port']),
        debug=int(arguments['--debug']),
        )


class PredictStream:
    """A class that helps make predictions through stdin and stdout.
    """
    def __init__(self):
        self.model = get_config()['model_persister'].read()
        self.predict_service = get_config()['predict_service']

    def process_line(self, line):
        predict_service = self.predict_service
        datas = ujson.loads(line)
        samples = [predict_service.sample_from_data(self.model, data)
                   for data in datas]
        samples = np.array(samples)
        params = predict_service.params_from_data(self.model, datas[0])
        return predict_service.predict(self.model, samples, **params)

    def listen(self, io_in, io_out, io_err):
        """Listens to provided io stream and writes predictions
        to output. In case of errors, the error stream will be used.
        """
        for line in io_in:
            if line.strip().lower() == 'exit':
                break

            try:
                y_pred = self.process_line(line)
            except Exception as e:
                io_out.write('[]\n')
                io_err.write(
                    "Error while processing input row: {}"
                    "{}: {}\n".format(line, type(e), e))
                io_err.flush()
            else:
                io_out.write(ujson.dumps(y_pred.tolist()))
                io_out.write('\n')
                io_out.flush()


def stream_cmd(argv=sys.argv[1:]):  # pragma: no cover
    """\
Start the streaming server, which listens to stdin, processes line
by line, and returns predictions.

The input should consist of a list of json objects, where each object
will result in a prediction.  Each line is processed in a batch.

Example input (must be on a single line):

  [{"sepal length": 1.0, "sepal width": 1.1, "petal length": 0.7,
    "petal width": 5}, {"sepal length": 1.0, "sepal width": 8.0,
    "petal length": 1.4, "petal width": 5}]

Example output:

  ["Iris-virginica","Iris-setosa"]

An input line with the word 'exit' will quit the streaming server.

Usage:
  pld-stream [options]

Options:
  -h --help                  Show this screen.
"""
    docopt(stream_cmd.__doc__, argv=argv)
    initialize_config()
    stream = PredictStream()
    stream.listen(sys.stdin, sys.stdout, sys.stderr)


@app.route('/list')
@PluggableDecorator('list_decorators')
@args_from_config
def list(model_persister):
    info = {
        'models': model_persister.list_models(),
        'properties': model_persister.list_properties(),
        }
    return make_ujson_response(info)


@PluggableDecorator('server_fit_decorators')
@args_from_config
def fit():
    param_converters = {
        'persist': lambda x: x.lower() in ('1', 't', 'true'),
        'activate': lambda x: x.lower() in ('1', 't', 'true'),
        'evaluate': lambda x: x.lower() in ('1', 't', 'true'),
        'persist_if_better_than': float,
        }
    params = {
        name: typ(request.form[name])
        for name, typ in param_converters.items()
        if name in request.form
        }
    thread, job_id = run_job(fit_base, **params)
    return make_ujson_response({'job_id': job_id}, status_code=200)


@PluggableDecorator('update_model_cache_decorators')
@args_from_config
def update_model_cache(model_persister):
    method = getattr(model_persister, 'update_cache', None)
    if method is not None:
        thread, job_id = run_job(model_persister.update_cache)
        return make_ujson_response({'job_id': job_id}, status_code=200)
    else:
        return make_ujson_response({}, status_code=503)


@PluggableDecorator('activate_decorators')
def activate():
    model_version = int(request.form['model_version'])
    try:
        activate_base(model_version=model_version)
    except LookupError:
        return make_ujson_response({}, status_code=503)
    else:
        return list()


def add_url_rule(rule, endpoint=None, view_func=None, app=app, **options):
    if isinstance(view_func, str):
        view_func = resolve_dotted_name(view_func)
    app.add_url_rule(rule, endpoint=endpoint, view_func=view_func, **options)



from datetime import datetime
import io
import json
import math
from threading import Thread
from time import sleep
from unittest.mock import call
from unittest.mock import Mock
from unittest.mock import patch

import dateutil.parser
from flask import request
import numpy as np
import pytest
import ujson
from werkzeug.exceptions import BadRequest


def dec(func):
    def inner(*args, **kwargs):
        """dec"""
        return func(*args, **kwargs) + '_decorated'
    return inner


class TestPredictService:
    @pytest.fixture
    def PredictService(self):
        from palladium.server import PredictService
        return PredictService

    def test_functional(self, PredictService, flask_app):
        model = Mock()
        model.threshold = 0.3
        model.size = 10
        # needed as hasattr would evaluate to True otherwise
        del model.threshold2
        del model.size2
        model.predict.return_value = np.array(['class1'])
        service = PredictService(
            mapping=[
                ('sepal length', 'float'),
                ('sepal width', 'float'),
                ('petal length', 'float'),
                ('petal width', 'float'),
                ('color', 'str'),
                ('age', 'int'),
                ('active', 'bool'),
                ('austrian', 'bool'),
                ],
            params=[
                ('threshold', 'float'),  # default will be overwritten
                ('size', 'int'),  # not provided, default value kept
                ('threshold2', 'float'),  # will be used, no default value
                ('size2', 'int'),  # not provided, no default value
                ])

        with flask_app.test_request_context():
            with patch('palladium.util.get_config') as get_config:
                get_config.return_value = {
                    'service_metadata': {
                        'service_name': 'iris',
                        'service_version': '0.1'
                    }
                }
                request = Mock(
                    args=dict([
                        ('sepal length', '5.2'),
                        ('sepal width', '3.5'),
                        ('petal length', '1.5'),
                        ('petal width', '0.2'),
                        ('color', 'purple'),
                        ('age', '1'),
                        ('active', 'True'),
                        ('austrian', 'False'),
                        ('threshold', '0.7'),
                        ('threshold2', '0.8'),
                        ]),
                    method='GET',
                    )
                resp = service(model, request)

        assert (model.predict.call_args[0][0] ==
                np.array([[5.2, 3.5, 1.5, 0.2,
                           'purple', 1, True, False]], dtype='object')).all()
        assert model.predict.call_args[1]['threshold'] == 0.7
        assert model.predict.call_args[1]['size'] == 10
        assert model.predict.call_args[1]['threshold2'] == 0.8
        assert 'size2' not in model.predict.call_args[1]
        assert resp.status_code == 200
        expected_resp_data = {
            "metadata": {
                "status": "OK",
                "error_code": 0,
                "service_name": "iris",
                "service_version": "0.1",
                },
            "result": "class1"
            }

        assert json.loads(resp.get_data(as_text=True)) == expected_resp_data

    def test_bad_request(self, PredictService, flask_app):
        predict_service = PredictService(mapping=[])
        model = Mock()
        request = Mock()
        with patch.object(predict_service, 'do') as psd:
            with flask_app.test_request_context():
                bad_request = BadRequest()
                bad_request.args = ('daniel',)
                psd.side_effect = bad_request
                resp = predict_service(model, request)
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 400
        assert resp_data == {
            "metadata": {
                "status": "ERROR",
                "error_code": -1,
                "error_message": "BadRequest: ('daniel',)"
                }
            }

    def test_predict_error(self, PredictService, flask_app):
        from palladium.interfaces import PredictError
        predict_service = PredictService(mapping=[])
        model = Mock()
        request = Mock()
        with patch.object(predict_service, 'do') as psd:
            with flask_app.test_request_context():
                psd.side_effect = PredictError("mymessage", 123)
                resp = predict_service(model, request)
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 500
        assert resp_data == {
            "metadata": {
                "status": "ERROR",
                "error_code": 123,
                "error_message": "mymessage",
                }
            }

    def test_generic_error(self, PredictService, flask_app):
        predict_service = PredictService(mapping=[])
        model = Mock()
        request = Mock()
        with patch.object(predict_service, 'do') as psd:
            with flask_app.test_request_context():
                psd.side_effect = KeyError("model")
                resp = predict_service(model, request)
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 500
        assert resp_data == {
            "metadata": {
                "status": "ERROR",
                "error_code": -1,
                "error_message": "KeyError: 'model'",
                }
            }

    def test_sample_from_data(self, PredictService):
        predict_service = PredictService(
            mapping=[
                ('name', 'str'),
                ('sepal width', 'int'),
                ],
            )

        model = Mock()
        request_args = {'name': 'myflower', 'sepal width': 3}
        sample = predict_service.sample_from_data(model, request_args)
        assert sample[0] == 'myflower'
        assert sample[1] == 3

    def test_unwrap_sample_get(self, PredictService, flask_app):
        predict_service = PredictService(
            mapping=[('text', 'str')],
            unwrap_sample=True,
            )
        model = Mock()
        model.predict.return_value = np.array([1])
        with flask_app.test_request_context():
            request = Mock(
                args=dict([
                    ('text', 'Hi this is text'),
                    ]),
                method='GET',
                )
            resp = predict_service(model, request)

        assert model.predict.call_args[0][0].ndim == 1
        model.predict.assert_called_with(np.array(['Hi this is text']))
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 200
        assert resp_data == {
            "metadata": {
                "status": "OK",
                "error_code": 0,
                },
            "result": 1,
            }

    def test_unwrap_sample_post(self, PredictService, flask_app):
        predict_service = PredictService(
            mapping=[('text', 'str')],
            unwrap_sample=True,
            )
        model = Mock()
        model.predict.return_value = np.array([1, 2])
        with flask_app.test_request_context():
            request = Mock(
                json=[
                    {'text': 'First piece of text'},
                    {'text': 'Second piece of text'},
                    ],
                method='POST',
                mimetype='application/json',
                )
            resp = predict_service(model, request)

        assert model.predict.call_args[0][0].ndim == 1
        assert (
            model.predict.call_args[0] ==
            np.array(['First piece of text', 'Second piece of text'])
            ).all()
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 200
        assert resp_data == {
            "metadata": {
                "status": "OK",
                "error_code": 0,
                },
            "result": [1, 2],
            }

    def test_probas(self, PredictService, flask_app):
        model = Mock()
        model.predict_proba.return_value = np.array([[0.1, 0.5, math.pi]])
        predict_service = PredictService(mapping=[], predict_proba=True)
        with flask_app.test_request_context():
            resp = predict_service(model, request)
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 200
        assert resp_data == {
            "metadata": {
                "status": "OK",
                "error_code": 0,
                },
            "result": [0.1, 0.5, math.pi],
            }

    def test_post_request(self, PredictService, flask_app):
        model = Mock()
        model.predict.return_value = np.array([3, 2])

        service = PredictService(
            mapping=[
                ('sepal length', 'float'),
                ('sepal width', 'float'),
                ('petal length', 'float'),
                ('petal width', 'float'),
                ],
            params=[
                ('threshold', 'float'),
                ],
            )

        request = Mock(
            json=[
                {
                    'sepal length': '5.2',
                    'sepal width': '3.5',
                    'petal length': '1.5',
                    'petal width': '0.2',
                    },
                {
                    'sepal length': '5.7',
                    'sepal width': '4.0',
                    'petal length': '2.0',
                    'petal width': '0.7',
                    },
                ],
            args=dict(threshold=1.0),
            method='POST',
            mimetype='application/json',
            )

        with flask_app.test_request_context():
            resp = service(model, request)

        assert (model.predict.call_args[0][0] == np.array([
            [5.2, 3.5, 1.5, 0.2],
            [5.7, 4.0, 2.0, 0.7],
            ],
            dtype='object',
            )).all()
        assert model.predict.call_args[1]['threshold'] == 1.0

        assert resp.status_code == 200
        expected_resp_data = {
            "metadata": {
                "status": "OK",
                "error_code": 0,
                },
            "result": [3, 2],
            }

        assert json.loads(resp.get_data(as_text=True)) == expected_resp_data

    @pytest.yield_fixture
    def mock_predict(self, monkeypatch):
        def mock_predict(model_persister, predict_service):
            return predict_service.entry_point
        monkeypatch.setattr(
            'palladium.server.predict', mock_predict)
        yield mock_predict

    def test_entry_point_not_set(
            self, config, flask_app_test, flask_client, mock_predict):
        from palladium.config import process_config

        config['model_persister'] = Mock()
        config['predict_service'] = {
            '!': 'palladium.server.PredictService',
            'mapping': [
                ('param', 'str'),
            ],
        }
        # set default predict_decorators
        config['predict_decorators'] = ['palladium.tests.test_server.dec']
        with flask_app_test.test_request_context():
            process_config(config)

        resp1 = flask_client.get(
            'predict?param=bla')

        # decorated result: default predict_decorators is defined
        assert resp1.get_data().decode('utf-8') == '/predict_decorated'

    def test_entry_point_multiple(
            self, config, flask_app_test, flask_client, mock_predict):
        from palladium.config import process_config

        config['model_persister'] = Mock()
        config['my_predict_service'] = {
            '!': 'palladium.server.PredictService',
            'mapping': [
                ('param', 'str'),
            ],
            'entry_point': '/predict1',
        }
        config['my_predict_service2'] = {
            '!': 'palladium.server.PredictService',
            'mapping': [
                ('param', 'str'),
            ],
            'entry_point': '/predict2',
            'decorator_list_name': 'predict_decorators2',
        }

        # only second predict service uses decorator list
        config['predict_decorators2'] = ['palladium.tests.test_server.dec']

        with flask_app_test.test_request_context():
            process_config(config)

        resp1 = flask_client.get(
            'predict1?param=bla')

        # no decorated result: default predict_decorators is not defined
        assert resp1.get_data().decode('utf-8') == '/predict1'

        resp2 = flask_client.get(
            'predict2?param=bla')

        # decorated result using predict_decorators2
        assert resp2.get_data().decode('utf-8') == '/predict2_decorated'

    def test_entry_point_multiple_conflict(
            self, config, flask_app_test, flask_client, mock_predict):
        from palladium.config import process_config

        config['model_persister'] = Mock()
        config['my_predict_service'] = {
            '!': 'palladium.server.PredictService',
            'mapping': [
                ('param', 'str'),
            ],
            'entry_point': '/predict1',  # <--
        }
        config['my_predict_service2'] = {
            '!': 'palladium.server.PredictService',
            'mapping': [
                ('param', 'str'),
            ],
            'entry_point': '/predict1',  # conflict: entry point exists
        }

        with pytest.raises(AssertionError):
            with flask_app_test.test_request_context():
                process_config(config)


class TestPredict:
    @pytest.fixture
    def predict(self):
        from palladium.server import predict
        return predict

    def test_predict_functional(self, config, flask_app_test, flask_client):
        from palladium.server import make_ujson_response
        model_persister = config['model_persister'] = Mock()
        predict_service = config['predict_service'] = Mock()
        with flask_app_test.test_request_context():
            from palladium.server import create_predict_function
            create_predict_function(
                '/predict', predict_service, 'predict_decorators', config)
            predict_service.return_value = make_ujson_response(
                'a', status_code=200)

        model = model_persister.read()

        resp = flask_client.get(
            'predict?sepal length=1.0&sepal width=1.1&'
            'petal length=0.777&petal width=5')

        resp_data = json.loads(resp.get_data(as_text=True))

        assert resp_data == 'a'
        assert resp.status_code == 200
        with flask_app_test.test_request_context():
            predict_service.assert_called_with(model, request)

    def test_unknown_exception(self, predict, flask_app):
        model_persister = Mock()
        model_persister.read.side_effect = KeyError('model')

        with flask_app.test_request_context():
            resp = predict(model_persister, Mock())
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 500
        assert resp_data == {
            "status": "ERROR",
            "error_code": -1,
            "error_message": "KeyError: 'model'",
            }


class TestAliveFunctional:
    def test_empty_process_state(self, config, flask_client):
        config['service_metadata'] = {'hello': 'world'}
        resp = flask_client.get('alive')
        assert resp.status_code == 200
        resp_data = json.loads(resp.get_data(as_text=True))

        assert sorted(resp_data.keys()) == ['memory_usage',
                                            'memory_usage_vms',
                                            'palladium_version',
                                            'process_metadata',
                                            'service_metadata']
        assert resp_data['service_metadata'] == config['service_metadata']

    def test_filled_process_state(self, config, process_store, flask_client):
        config['alive'] = {'process_store_required': ('model', 'data')}

        before = datetime.now()
        process_store['model'] = Mock(__metadata__={'hello': 'is it me'})
        process_store['data'] = Mock(__metadata__={'bye': 'not you'})
        after = datetime.now()

        resp = flask_client.get('alive')
        assert resp.status_code == 200
        resp_data = json.loads(resp.get_data(as_text=True))

        model_updated = dateutil.parser.parse(resp_data['model']['updated'])
        data_updated = dateutil.parser.parse(resp_data['data']['updated'])
        assert before < model_updated < after
        assert resp_data['model']['metadata'] == {'hello': 'is it me'}
        assert before < data_updated < after
        assert resp_data['data']['metadata'] == {'bye': 'not you'}

    def test_missing_process_state(self, config, process_store, flask_client):
        config['alive'] = {'process_store_required': ('model', 'data')}
        process_store['model'] = Mock(__metadata__={'hello': 'is it me'})

        resp = flask_client.get('alive')
        assert resp.status_code == 503
        resp_data = json.loads(resp.get_data(as_text=True))

        assert resp_data['model']['metadata'] == {'hello': 'is it me'}
        assert resp_data['data'] == 'N/A'


class TestPredictStream:
    @pytest.fixture
    def PredictStream(self):
        from palladium.server import PredictStream
        return PredictStream

    @pytest.fixture
    def stream(self, config, PredictStream):
        config['model_persister'] = Mock()
        predict_service = config['predict_service'] = Mock()
        predict_service.sample_from_data.side_effect = (
            lambda model, data: data)
        predict_service.params_from_data.side_effect = (
            lambda model, data: data)
        return PredictStream()

    def test_listen_direct_exit(self, stream):
        io_in = io.StringIO()
        io_out = io.StringIO()
        io_err = io.StringIO()

        stream_thread = Thread(
            target=stream.listen(io_in, io_out, io_err))
        stream_thread.start()
        io_in.write('EXIT\n')
        stream_thread.join()
        io_out.seek(0)
        io_err.seek(0)
        assert len(io_out.read()) == 0
        assert len(io_err.read()) == 0
        assert stream.predict_service.predict.call_count == 0

    def test_listen(self, stream):
        io_in = io.StringIO()
        io_out = io.StringIO()
        io_err = io.StringIO()
        lines = [
            '[{"id": 1, "color": "blue", "length": 1.0}]\n',
            '[{"id": 1, "color": "{\\"a\\": 1, \\"b\\": 2}", "length": 1.0}]\n',
            '[{"id": 1, "color": "blue", "length": 1.0}, {"id": 2, "color": "{\\"a\\": 1, \\"b\\": 2}", "length": 1.0}]\n',
        ]
        for line in lines:
            io_in.write(line)

        io_in.write('EXIT\n')
        io_in.seek(0)
        predict = stream.predict_service.predict
        predict.side_effect = (
            lambda model, samples, **params:
            np.array([{'result': 1}] * len(samples))
            )
        stream_thread = Thread(
            target=stream.listen(io_in, io_out, io_err))
        stream_thread.start()
        stream_thread.join()
        io_out.seek(0)
        io_err.seek(0)
        assert len(io_err.read()) == 0
        assert io_out.read() == (
            ('[{"result":1}]\n' * 2) + ('[{"result":1},{"result":1}]\n'))
        assert predict.call_count == 3
        # check if the correct arguments are passed to predict call
        assert predict.call_args_list[0][0][1] == np.array([
            {'id': 1, 'color': 'blue', 'length': 1.0}])
        assert predict.call_args_list[1][0][1] == np.array([
            {'id': 1, 'color': '{"a": 1, "b": 2}', 'length': 1.0}])
        assert (predict.call_args_list[2][0][1] == np.array([
            {'id': 1, 'color': 'blue', 'length': 1.0},
            {'id': 2, 'color': '{"a": 1, "b": 2}', 'length': 1.0},
            ])).all()

        # check if string representation of attribute can be converted to json
        assert ujson.loads(predict.call_args_list[1][0][1][0]['color']) == {
            "a": 1, "b": 2}

    def test_predict_error(self, stream):
        from palladium.interfaces import PredictError

        io_in = io.StringIO()
        io_out = io.StringIO()
        io_err = io.StringIO()

        line = '[{"hey": "1"}]\n'
        io_in.write(line)
        io_in.write('EXIT\n')
        io_in.seek(0)
        stream.predict_service.predict.side_effect = PredictError('error')

        stream_thread = Thread(
            target=stream.listen(io_in, io_out, io_err))
        stream_thread.start()
        stream_thread.join()

        io_out.seek(0)
        io_err.seek(0)
        assert io_out.read() == '[]\n'
        assert io_err.read() == (
            "Error while processing input row: {}"
            "<class 'palladium.interfaces.PredictError'>: "
            "error (-1)\n".format(line))
        assert stream.predict_service.predict.call_count == 1

    def test_predict_params(self, config, stream):
        from palladium.server import PredictService
        line = '[{"length": 1.0, "width": 1.0, "turbo": "true"}]'

        model = Mock()
        model.predict.return_value = np.array([[{'class': 'a'}]])
        model.turbo = False
        model.magic = False
        stream.model = model

        mapping = [
            ('length', 'float'),
            ('width', 'float'),
            ]
        params = [
            ('turbo', 'bool'),  # will be set by request args
            ('magic', 'bool'),  # default value will be used
            ]
        stream.predict_service = PredictService(
            mapping=mapping,
            params=params,
        )

        expected = [{'class': 'a'}]
        result = stream.process_line(line)
        assert result == expected
        assert model.predict.call_count == 1
        assert (model.predict.call_args[0][0] == np.array([[1.0, 1.0]])).all()
        assert model.predict.call_args[1]['turbo'] is True
        assert model.predict.call_args[1]['magic'] is False


class TestList:
    @pytest.fixture
    def list(self):
        from palladium.server import list
        return list

    def test_it(self, config, process_store, flask_client):
        mp = config['model_persister'] = Mock()
        mp.list_models.return_value = ['one', 'two']
        mp.list_properties.return_value = {'hey': 'there'}
        resp = flask_client.get('list')
        assert resp.status_code == 200
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp_data == {
            'models': ['one', 'two'],
            'properties': {'hey': 'there'},
            }


class TestFitFunctional:
    @pytest.fixture
    def fit(self):
        from palladium.server import fit
        return fit

    @pytest.fixture
    def jobs(self, process_store):
        jobs = process_store['process_metadata'].setdefault('jobs', {})
        yield jobs
        jobs.clear()

    def test_it(self, fit, config, jobs, flask_app):
        dsl, model, model_persister = Mock(), Mock(), Mock()
        del model.cv_results_
        X, y = Mock(), Mock()
        dsl.return_value = X, y
        config['dataset_loader_train'] = dsl
        config['model'] = model
        config['model_persister'] = model_persister
        with flask_app.test_request_context(method='POST'):
            resp = fit()
        sleep(0.05)
        resp_json = json.loads(resp.get_data(as_text=True))
        job = jobs[resp_json['job_id']]
        assert job['status'] == 'finished'
        assert job['info'] == str(model)

    @pytest.mark.parametrize('args, args_expected', [
        (
            {'persist': '1', 'activate': '0', 'evaluate': 't'},
            {'persist': True, 'activate': False, 'evaluate': True},
        ),
        (
            {'persist_if_better_than': '0.234'},
            {'persist_if_better_than': 0.234},
        ),
    ])
    def test_pass_args(self, fit, flask_app, args, args_expected):
        with patch('palladium.server.fit_base') as fit_base:
            fit_base.__name__ = 'mock'
            with flask_app.test_request_context(method='POST', data=args):
                fit()
            sleep(0.02)
        assert fit_base.call_args == call(**args_expected)


class TestUpdateModelCacheFunctional:
    @pytest.fixture
    def update_model_cache(self):
        from palladium.server import update_model_cache
        return update_model_cache

    @pytest.fixture
    def jobs(self, process_store):
        jobs = process_store['process_metadata'].setdefault('jobs', {})
        yield jobs
        jobs.clear()

    def test_success(self, update_model_cache, config, jobs, flask_app):
        model_persister = Mock()
        config['model_persister'] = model_persister
        with flask_app.test_request_context(method='POST'):
            resp = update_model_cache()
        sleep(0.02)
        resp_json = json.loads(resp.get_data(as_text=True))
        job = jobs[resp_json['job_id']]
        assert job['status'] == 'finished'
        assert job['info'] == repr(model_persister.update_cache())

    def test_unavailable(self, update_model_cache, config, jobs, flask_app):
        model_persister = Mock()
        del model_persister.update_cache
        config['model_persister'] = model_persister
        with flask_app.test_request_context(method='POST'):
            resp = update_model_cache()
        assert resp.status_code == 503


class TestActivateFunctional:
    @pytest.fixture
    def activate(self):
        from palladium.server import activate
        return activate

    @pytest.fixture
    def activate_base_mock(self, monkeypatch):
        func = Mock()
        monkeypatch.setattr('palladium.server.activate_base', func)
        return func

    def test_success(self, activate, activate_base_mock, config, flask_app):
        model_persister = Mock(
            list_models=lambda: {'be': 'first'},
            list_properties=lambda: {'be': 'twice'},
            )
        config['model_persister'] = model_persister
        with flask_app.test_request_context(
            method='POST',
            data={'model_version': 123},
        ):
            resp = activate()
        assert resp.status_code == 200
        assert resp.json == {
            'models': {'be': 'first'},
            'properties': {'be': 'twice'},
            }

    def test_lookuperror(self, activate, activate_base_mock, flask_app):
        activate_base_mock.side_effect = LookupError
        with flask_app.test_request_context(
            method='POST',
            data={'model_version': 123},
        ):
            resp = activate()
        assert resp.status_code == 503


def _test_add_url_rule_func():
    return b'A OK'


class TestAddUrlRule:
    @pytest.fixture
    def add_url_rule(self):
        from palladium.server import add_url_rule
        return add_url_rule

    def test_it(self, add_url_rule, flask_client):
        add_url_rule(
            '/okay',
            view_func='palladium.tests.test_server._test_add_url_rule_func',
            )
        resp = flask_client.get('/okay')
        assert resp.data == b'A OK'



from copy import deepcopy
import pytest


@pytest.fixture
def flask_client(flask_app):
    return flask_app.test_client()


@pytest.fixture
def process_store(request):
    from palladium.util import process_store
    orig = deepcopy(process_store)
    yield process_store
    process_store.clear()
    process_store.update(orig)



""":class:`~palladium.interfaces.DatasetLoader` implementations.
"""

import pandas.io.parsers
import pandas.io.sql
from sklearn.datasets import fetch_openml
from sqlalchemy import create_engine

from .interfaces import DatasetLoader
from .util import logger
from .util import PluggableDecorator
from .util import process_store
from .util import RruleThread


class CSV(DatasetLoader):
    """A :class:`~palladium.interfaces.DatasetLoader` that uses
    :func:`pandas.io.parsers.read_csv` to load data from a file or
    URL.
    """
    pandas_read = staticmethod(pandas.io.parsers.read_csv)

    def __init__(self, path, target_column=None,
                 ndarray=True, **kwargs):
        """
        :param str path:
          The *path* represents a filesystem path or URL that's passed
          on as the *filepath_or_buffer* argument to
          :func:`read_table`.

        :param str target_column:
          The column in the table to load that represents the target
          value.  This column will not be part of the returned *data*.

          If *target_column* is None, then the target return value
          will be None as well.

        :param kwargs:
          All other keyword parameters are passed on to
          :func:`pandas.io.parsers.read_table`.  The most useful
          options may be *usecols* to select which columns of the
          table to use, *skiprows* to skip a certain number of rows at
          the beginning and *nrows* to select a given number of rows
          only.
        """
        self.path = path
        self.target_column = target_column
        self.ndarray = ndarray
        self.kwargs = kwargs

    def __call__(self):
        """See :meth:`palladium.interfaces.DatasetLoader.__call__`.
        """
        df = self.pandas_read(self.path, **self.kwargs)
        data_columns = [col for col in df.columns if col != self.target_column]
        data = df[data_columns]
        target = None
        if self.target_column:
            target = df[self.target_column]
        if self.ndarray:
            return data.values, target.values if target is not None else None
        else:
            return data, target


class Table(CSV):
    """A :class:`~palladium.interfaces.DatasetLoader` that uses the
    deprecated :func:`pandas.io.parsers.read_table` to load data from
    a file or URL.
    """
    pandas_read = staticmethod(pandas.io.parsers.read_table)


class SQL(DatasetLoader):
    """A :class:`~palladium.interfaces.DatasetLoader` that uses
    :func:`pandas.io.sql.read_sql` to load data from an SQL database.
    Supports all databases that SQLAlchemy has support for.
    """
    pandas_read = staticmethod(pandas.io.sql.read_sql)

    def __init__(self, url, sql, target_column=None, ndarray=True, **kwargs):
        """
        :param str url:
          The database *url* that'll be used to make a connection.
          Format follows RFC-1738.

        :param str sql:
          SQL query to be executed or database table name.

        :param str target_column:
          The name of the column used as the target.  (All other
          columns are considered feature data.)

        :param kwargs:
          All other keyword parameters are passed on to
          :func:`pandas.io.parsers.read_sql`.
        """
        self.engine = create_engine(url)
        self.sql = sql
        self.target_column = target_column
        self.ndarray = ndarray
        self.kwargs = kwargs

    def __call__(self):
        """See :meth:`palladium.interfaces.DatasetLoader.__call__`.
        """
        df = self.pandas_read(self.sql, self.engine, **self.kwargs)
        data_columns = [col for col in df.columns if col != self.target_column]
        data = df[data_columns]
        target = None
        if self.target_column:
            target = df[self.target_column]
        if self.ndarray:
            return data.values, target.values if target is not None else None
        else:
            return data, target


class OpenML(DatasetLoader):  # pragma: no cover
    """A :class:`~palladium.interfaces.DatasetLoader` that uses
    scikit-learn's :func:`sklearn.datasets.fetch_openml` to load data
    from OpenML.
    """
    def __init__(self, name):
        """
        :param str name:
          The dataset name from OpenML.
          Examples: "wine-quality-red", "diabetes"
        """
        self.name = name

    def __call__(self):
        dataset = fetch_openml(self.name)
        return dataset.data, dataset.target


class EmptyDatasetLoader(DatasetLoader):
    """This :class:`~palladium.interfaces.DatasetLoader` can be used if no
    actual data should be loaded.  Returns a ``(None, None)`` tuple.
    """
    def __call__(self):
        return None, None


class ScheduledDatasetLoader(DatasetLoader):
    """A :class:`~palladium.interfaces.DatasetLoader` that loads
    periodically data into RAM to make it available to the prediction
    server inside the ``process_store``.

    :class:`~ScheduledDatasetLoader` wraps another
    :class:`~palladium.interfaces.DatasetLoader` class that it uses to do
    the actual loading of the data.

    An *update_cache_rrule* is used to define how often data should be
    loaded anew.

    This class' :meth:`~ScheduledDatasetLoader.read` read method never
    calls the underlying dataset loader.  It will only ever fetch the
    data from the in-memory cache.
    """
    cache = process_store
    key = 'data'

    def __init__(self,
                 impl,
                 update_cache_rrule,
                 ):
        """
        :param palladium.interfaces.DatasetLoader impl:
          The underlying (decorated) dataset loader object.

        :param dict update_cache_rrule:
          Keyword arguments for a :class:`dateutil.rrule.rrule` that
          determines when the cache will be updated.  See
          :class:`~palladium.util.RruleThread` for details.
        """
        self.impl = impl
        self.update_cache_rrule = update_cache_rrule

    def initialize_component(self, config):
        self.update_cache()
        logger.info("{}: initial fill of cache done.".format(
            self.__class__.__name__))

        self.thread = RruleThread(
            func=self.update_cache, rrule=self.update_cache_rrule)
        self.thread.start()

    def __call__(self):
        return self.cache[self.key]

    @PluggableDecorator('update_data_decorators')
    def update_cache(self, *args, **kwargs):
        data = self.impl()
        self.cache[self.key] = data
        return data



from collections import OrderedDict
from datetime import datetime
import threading
from time import sleep
from unittest.mock import Mock
from unittest.mock import patch

from dateutil import rrule
import pytest


class TestResolveDottedName:
    @pytest.fixture
    def resolve_dotted_name(self):
        from palladium.util import resolve_dotted_name
        return resolve_dotted_name

    def test_with_colon(self, resolve_dotted_name):
        dotted = 'palladium.tests.test_util:TestResolveDottedName.test_with_colon'
        assert (resolve_dotted_name(dotted) is
                TestResolveDottedName.test_with_colon)

    def test_with_dots(self, resolve_dotted_name):
        dotted = 'palladium.tests.test_util.TestResolveDottedName'
        assert (resolve_dotted_name(dotted) is
                TestResolveDottedName)

    def test_module(self, resolve_dotted_name):
        resolve_dotted_name('threading') is threading


def test_args_from_config(config):
    from palladium.util import args_from_config

    args = []
    config.update({
        'arg1': 'config1',
        'arg2': 'config2',
        'arg4': 'config4',
        })

    @args_from_config
    def myfunc(arg1, arg2, arg3, arg4='default4', arg5='default5'):
        args.append((arg1, arg2, arg3, arg4, arg5))

    myfunc(arg3='myarg3')
    assert args[-1] == ('config1', 'config2', 'myarg3', 'config4', 'default5')

    myfunc(arg2='myarg2', arg3='myarg3', arg5='myarg5')
    assert args[-1] == ('config1', 'myarg2', 'myarg3', 'config4', 'myarg5')

    myfunc('myarg1', arg3='myarg3')
    assert args[-1] == ('myarg1', 'config2', 'myarg3', 'config4', 'default5')

    with pytest.raises(TypeError) as e:
        myfunc()
    assert "Maybe you forgot to set" in str(e.value)


class TestProcessStore:
    @pytest.fixture
    def store(self):
        from palladium.util import ProcessStore
        return ProcessStore()

    def test_get(self, store):
        assert store.get('somekey') is None
        store['somekey'] = '1'
        assert store.get('somekey') == '1'

    def test_getitem(self, store):
        with pytest.raises(KeyError):
            store['somekey']
        store['somekey'] = '1'
        assert store['somekey'] == '1'

    def test_in(self, store):
        assert not 'somekey' in store
        store['somekey'] = '1'
        assert 'somekey' in store

    def test_setitem(self, store):
        store['somekey'] = '1'
        assert store['somekey'] == '1'
        store['somekey'] = '2'
        assert store['somekey'] == '2'

    def test_delitem(self, store):
        store['somekey'] = '1'
        del store['somekey']
        with pytest.raises(KeyError):
            store['somekey']

    def test_mtime_no_entry(self, store):
        with pytest.raises(KeyError):
            assert store.mtime['somekey']

    def test_mtime_setitem(self, store):
        dt0 = datetime.now()
        store['somekey'] = '1'
        sleep(0.005)  # make sure that we're not too fast
        dt1 = datetime.now()
        assert dt0 < store.mtime['somekey'] < dt1
        store['somekey'] = '2'
        sleep(0.005)  # make sure that we're not too fast
        dt2 = datetime.now()
        assert dt1 < store.mtime['somekey'] < dt2

    def test_mtime_delitem(self, store):
        store['somekey'] = '1'
        del store['somekey']
        with pytest.raises(KeyError):
            store.mtime['somekey']

    def test_init(self, store):
        store = store.__class__({'somekey': '1'})
        assert store['somekey'] == '1'
        assert 'somekey' in store.mtime

    def test_clear(self, store):
        store['somekey'] = '1'
        store.clear()
        assert 'somekey' not in store
        assert 'somekey' not in store.mtime


class TestRruleThread:
    @pytest.fixture
    def RruleThread(self):
        from palladium.util import RruleThread
        return RruleThread

    def test_rrule_from_dict(self, RruleThread):
        func = Mock()
        now = datetime.now()
        rrule_info = {
            'freq': 'DAILY',
            'dtstart': '2014-10-30T13:21:18',
            }
        expected = rrule.rrule(
            rrule.DAILY, dtstart=datetime(2014, 10, 30, 13, 21, 18))

        thread = RruleThread(func, rrule_info)
        assert thread.rrule.after(now) == expected.after(now)

    def test_last_execution(self, RruleThread):
        func = Mock()
        thread = RruleThread(func, rrule.rrule(
            rrule.MONTHLY,
            bymonthday=1,
            dtstart=datetime(2014, 10, 30, 13, 21, 18)))
        thread.last_execution = datetime(2014, 10, 30, 13, 21, 18)
        thread.start()
        sleep(0.02)
        assert func.call_count == 1

    def test_func_raises(self, RruleThread):
        func = Mock(__name__='buggy')
        func.side_effect = KeyError
        thread = RruleThread(func, rrule.rrule(
            rrule.MONTHLY,
            bymonthday=1,
            dtstart=datetime(2014, 10, 30, 13, 21, 18)))
        thread.last_execution = datetime(2014, 10, 30, 13, 21, 18)

        with patch('palladium.util.logger') as logger:
            thread.start()
            sleep(0.02)
            assert func.call_count == 1
            assert logger.exception.call_count == 1

    def test_sleep_between_checks(self, RruleThread):
        func = Mock()
        rr = Mock()
        rr.between.return_value = False
        thread = RruleThread(func, rr, sleep_between_checks=0.0010)
        thread.start()
        sleep(0.02)
        assert func.call_count == 0
        assert rr.between.call_count > 1


class TestUpgrade:
    @pytest.fixture
    def upgrade(self):
        from palladium.util import upgrade
        return upgrade

    def test_no_args(self, upgrade):
        persister = Mock()
        upgrade(persister)
        persister.upgrade.assert_called_with(from_version=None)

    def test_from_version(self, upgrade):
        persister = Mock()
        upgrade(persister, from_version='0.1')
        persister.upgrade.assert_called_with(from_version='0.1')

    def test_from_version_and_to_version(self, upgrade):
        persister = Mock()
        upgrade(persister, from_version='0.1', to_version='0.2')
        persister.upgrade.assert_called_with(from_version='0.1',
                                             to_version='0.2')


class TestExport:
    @pytest.fixture
    def export(self):
        from palladium.util import export
        return export

    def test_no_args(self, export):
        persister = Mock()
        persister_export = Mock()
        export(persister, persister_export)
        persister.read.assert_called_with(None)
        persister_export.write.assert_called_with(persister.read())
        persister_export.activate.assert_called_with(persister_export.write())

    def test_model_version(self, export):
        persister = Mock()
        export(persister, Mock(), model_version=123)
        persister.read.assert_called_with(123)

    def test_activate(self, export):
        persister_export = Mock()
        export(Mock(), persister_export, activate=False)
        assert persister_export.activate.call_count == 0


def dec1(func):
    def inner(a, b):
        """dec1"""
        return func(a, b) * 10
    return inner


def dec2(func):
    def inner(a, b):
        """dec2"""
        return func(a, b) + 2
    return inner


class TestPluggableDecorator:
    @pytest.fixture
    def PluggableDecorator(self):
        from palladium.util import PluggableDecorator
        return PluggableDecorator

    def test_no_config(self, PluggableDecorator, config):
        decorator = PluggableDecorator('decorator_list')

        @decorator
        def myfunc(a, b):
            """myfunc"""
            return a + b

        result = myfunc(2, 3)
        assert result == 5
        assert myfunc.__doc__ == "myfunc"

    def test_two_decorators(self, PluggableDecorator, config):
        config['decorator_list'] = [
            'palladium.tests.test_util.dec1',
            'palladium.tests.test_util.dec2']
        decorator = PluggableDecorator('decorator_list')

        @decorator
        def myfunc(a, b):
            """myfunc"""
            return a + b

        result = myfunc(2, 3)
        assert result == 52
        assert myfunc.__doc__ == "myfunc"

    def test_wrapped_empty_list(self, PluggableDecorator, config):
        def myfunc(a, b):
            """myfunc"""
            return a + b

        myfunc2 = PluggableDecorator('decorator_list')(myfunc)
        assert hasattr(myfunc, '__wrapped__') is False
        assert myfunc2.__wrapped__ is myfunc

    def test_wrapped(self, PluggableDecorator, config):
        config['decorator_list'] = [
            'palladium.tests.test_util.dec1',
            'palladium.tests.test_util.dec2']

        def myfunc(a, b):
            """myfunc"""
            return a + b

        myfunc2 = PluggableDecorator('decorator_list')(myfunc)
        assert hasattr(myfunc, '__wrapped__') is False
        assert hasattr(myfunc2, '__wrapped__') is True
        assert myfunc2.__wrapped__ is myfunc


class TestSessionScope:

    def test_success(self):
        from palladium.util import session_scope
        session = Mock()
        with session_scope(session):
            session.query()
        assert session.commit.call_count == 1
        assert session.rollback.call_count == 0
        assert session.close.call_count == 1

    def test_exception(self):
        from palladium.util import session_scope
        session = Mock()
        with pytest.raises(KeyError):
            with session_scope(session):
                raise KeyError('Error')
        assert session.commit.call_count == 0
        assert session.rollback.call_count == 1
        assert session.close.call_count == 1


class TestApplyKwargs:
    @pytest.fixture
    def apply_kwargs(self):
        from palladium.util import apply_kwargs
        return apply_kwargs

    @pytest.fixture
    def myfunc(self):
        calls = []

        def myfunc(one, two, three='three'):
            calls.append((one, two, three))

        myfunc.calls = calls
        return myfunc

    def test_it(self, apply_kwargs, myfunc):
        apply_kwargs(myfunc, one=1, two='two', four=4)
        assert myfunc.calls == [(1, 'two', 'three')]

    def test_typeerror(self, apply_kwargs, myfunc):
        with pytest.raises(TypeError):
            apply_kwargs(myfunc, two='two', four=4)


class TestPartial:
    calls = []

    @classmethod
    def myfunc(cls, one, two, three='three'):
        cls.calls.append((one, two, three))

    @pytest.fixture
    def Partial(self):
        from palladium.util import Partial
        return Partial

    def test_it(self, Partial):
        Partial(self.myfunc, two=2)(one='one')
        assert self.calls == [('one', 2, 'three')]
        self.__class__.calls = []

    def test_dotted(self, Partial):
        Partial('palladium.tests.test_util:TestPartial.myfunc', two=2)(one='one')
        assert self.calls == [('one', 2, 'three')]
        self.__class__.calls = []


class TestRunJob:
    @pytest.fixture
    def run_job(self):
        from palladium.util import run_job
        return run_job

    @pytest.fixture
    def jobs(self, process_store):
        jobs = process_store['process_metadata']['jobs'] = {}
        yield jobs
        jobs.clear()

    def test_simple(self, run_job, jobs):
        def myfunc(add):
            nonlocal result
            result += add
            return add

        result = 0
        results = []
        for i in range(3):
            results.append(run_job(myfunc, add=i))
        sleep(0.02)
        assert result == 3
        assert len(jobs) == len(results) == 3
        assert set(jobs.keys()) == set(r[1] for r in results)
        assert all(j['status'] == 'finished' for j in jobs.values())
        assert set(j['info'] for j in jobs.values()) == set(['0', '1', '2'])

    def test_exception(self, run_job, jobs):
        def myfunc(divisor):
            nonlocal result
            result /= divisor

        result = 1
        num_threads_before = len(threading.enumerate())
        for i in range(3):
            run_job(myfunc, divisor=i)
        sleep(0.02)
        num_threads_after = len(threading.enumerate())

        assert num_threads_before == num_threads_after
        assert result == 0.5
        job1, job2, job3 = sorted(jobs.values(), key=lambda x: x['started'])
        assert job1['status'] == 'error'
        assert 'division by zero' in job1['info']
        assert job2['status'] == 'finished'
        assert job3['status'] == 'finished'

    def test_lifecycle(self, run_job, jobs):
        def myfunc(tts):
            sleep(tts)

        num_threads_before = len(threading.enumerate())
        for i in range(3):
            run_job(myfunc, tts=5*i/100)

        job1, job2, job3 = sorted(jobs.values(), key=lambda x: x['started'])

        samples = []
        for i in range(50):
            samples.append((
                job1['status'],
                job2['status'],
                job3['status'],
                len(threading.enumerate()),
                ))
            sleep(1/100)

        got = list(OrderedDict.fromkeys(samples))
        expected = [
            ('finished', 'running', 'running', num_threads_before+2),
            ('finished', 'finished', 'running', num_threads_before+1),
            ('finished', 'finished', 'finished', num_threads_before+0),
        ]
        assert got == expected



from unittest.mock import Mock
from unittest.mock import call
from unittest.mock import patch

import pytest


class TestTest:
    @pytest.fixture
    def test(self):
        from palladium.eval import test
        return test

    def test_test(self, test):
        dataset_loader_test, model_persister = Mock(), Mock()
        X, y = object(), object()
        dataset_loader_test.return_value = X, y
        model = model_persister.read.return_value
        model.__metadata__ = {'version': 77}

        test(dataset_loader_test, model_persister, model_version=77)

        dataset_loader_test.assert_called_with()
        model_persister.read.assert_called_with(version=77)
        model.score.assert_called_with(X, y)

    def test_test_no_score(self, test):
        dataset_loader_test, model_persister = Mock(), Mock()
        X, y = object(), object()
        dataset_loader_test.return_value = X, y
        model_persister.read.return_value = Mock(spec=['fit', 'predict'])
        model_persister.read.return_value.__metadata__ = {'version': 99}

        with pytest.raises(ValueError):
            test(dataset_loader_test, model_persister)

    def test_test_two_scores(self, test):
        dataset_loader_test, model_persister, scorer = Mock(), Mock(), Mock()
        X, y = object(), object()
        dataset_loader_test.return_value = X, y
        model = model_persister.read.return_value
        model.__metadata__ = {'version': 77}

        test(dataset_loader_test, model_persister,
             scoring=scorer, model_version=77)

        dataset_loader_test.assert_called_with()
        model_persister.read.assert_called_with(version=77)
        scorer.assert_called_with(model, X, y)
        assert model.score.call_count == 0

    def test_scoring_dict(self, test):
        dataset_loader_test, model_persister = Mock(), Mock()
        scoring = {'AUC': Mock(), 'accuracy': Mock()}
        scoring['AUC'].return_value = 1.23
        scoring['accuracy'].return_value = 3.45
        X, y = object(), object()
        dataset_loader_test.return_value = X, y
        model = model_persister.read.return_value
        model.__metadata__ = {'version': 77}

        result = test(dataset_loader_test, model_persister,
                      scoring=scoring, model_version=77)

        scoring['AUC'].assert_called_with(model, X, y)
        scoring['accuracy'].assert_called_with(model, X, y)
        assert sorted(result) == ['AUC: 1.23', 'accuracy: 3.45']


class TestList:
    @pytest.fixture
    def list(self):
        from palladium.eval import list
        return list

    def test(self, list):
        model_persister = Mock()
        model_persister.list_models.return_value = [{1: 2}]
        model_persister.list_properties.return_value = {5: 6}
        with patch('palladium.eval.pprint') as pprint:
            list(model_persister)
        assert pprint.mock_calls[0] == call([{1: 2}])
        assert pprint.mock_calls[1] == call({5: 6})



"""Interfaces defining the behaviour of Palladium's components.
"""

from abc import abstractmethod
from abc import ABCMeta

from sklearn.base import BaseEstimator

from . import __version__
from .util import PluggableDecorator


def annotate(obj, metadata=None):
    base_metadata = getattr(obj, '__metadata__', {})
    if metadata is not None:
        base_metadata.update(metadata)
        obj.__metadata__ = base_metadata
    return obj.__metadata__


class DatasetLoaderMeta(ABCMeta):
    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs, **kwargs)
        cls.__call__ = PluggableDecorator('load_data_decorators')(cls.__call__)


class DatasetLoader(metaclass=DatasetLoaderMeta):
    """A :class:`~palladium.interfaces.DatasetLoader` is responsible for
    loading datasets for use in training and evaluation.
    """

    @abstractmethod
    def __call__(self):
        """Loads the data and returns a tuple *(data, target)*, or
        *(X, y)*.

        :return:
          A tuple *(data, target*).

          *data* is a two dimensional numpy array with shape n x m
          (one row per example).

          *target* is a one dimensional array with n target values.

          *target* may be ``None`` if there is no target value,
          e.g. in an unsupervised learning setting.

        :rtype: tuple
        """


class CrossValidationGenerator(metaclass=ABCMeta):
    """A :class:`CrossValidationGenerator` provides train/test indices
    to split data in train and validation sets.

    :class:`CrossValidationGenerator` corresponds to the *cross
    validation generator* interface of scikit-learn.
    """

    @abstractmethod
    def __iter__(self):
        """
        :return:
          Tuples of train/test indices.
        """


class Model(BaseEstimator, metaclass=ABCMeta):
    """A :class:`Model` can be :meth:`~Model.fit` to data and can be
    used to :meth:`~Model.predict` data.

    :class:`Model` corresponds to the *estimators* interface of
    scikit-learn.
    """

    @abstractmethod
    def fit(self, X, y=None):
        """Fit to data array *X* and possibly a target array *y*.

        :return: self
        """

    @abstractmethod
    def predict(self, X, **kw):
        """Predict classes for data array *X* with shape n x m.

        Some models may accept additional keyword arguments.

        :return:
          A numpy array of length n with the predicted classes (for
          classification problems) or numeric values (for regression
          problems).

        :raises:
          May raise a :class:`PredictError` to indicate that some
          condition made it impossible to deliver a prediction.
        """

    def predict_proba(self, X, **kw):
        """Predict probabilities for data array *X* with shape n x m.

        :return:
          A numpy array of length n x c with a list class
          probabilities per sample.

        :raises:
          :class:`NotImplementedError` if not applicable.
        """
        raise NotImplementedError()  # pragma: no cover


class PredictError(Exception):
    """Raised by :meth:`Model.predict` to indicate that some condition
    made it impossible to deliver a prediction.
    """
    def __init__(self, error_message, error_code=-1):
        self.error_message = error_message
        self.error_code = error_code

    def __str__(self):
        return "{} ({})".format(self.error_message, self.error_code)


class ModelPersisterMeta(ABCMeta):
    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs, **kwargs)
        cls.read = PluggableDecorator('read_model_decorators')(cls.read)
        cls.write = PluggableDecorator('write_model_decorators')(cls.write)


class ModelPersister(metaclass=ModelPersisterMeta):
    @abstractmethod
    def read(self, version=None):
        """Returns a :class:`Model` instance.

        :param str version:
          *version* may be used to read a specific version of a model.
          If *version* is ``None``, returns the active model.

        :return:
          The model object.

        :raises:
          LookupError if no model was available.
        """

    @abstractmethod
    def write(self, model):
        """Persists a :class:`Model` and returns a new version number.

        It is the :class:`ModelPersister`'s responsibility to annotate
        the 'version' information onto the model before it is saved.

        The new model will initially be inactive.  Use
        :meth:`ModelPersister.activate` to activate the model.

        :return:
          The new model's version identifier.
        """

    @abstractmethod
    def activate(self, version):
        """Set the model with the given *version* to be the active
        one.

        Implies that any previously active model becomes inactive.

        :param str version:
          The *version* of the model that's activated.

        :raises:
          LookupError if no model with given *version* exists.
        """

    def delete(self, version):
        """Delete the model with the given *version* from the
        database.

        :param str version:
          The *version* of the model that's activated.

        :raises:
          LookupError if no model with given *version* exists.
        """

    @abstractmethod
    def list_models(self):
        """List metadata of all available models.

        :return:
          A list of dicts, with each dict containing information about
          one of the available models.  Each dict is guaranteed to
          contain the ``version`` key, which is the same version
          number that :meth:`ModelPersister.read` accepts for loading
          specific models.
        """

    @abstractmethod
    def list_properties(self):
        """List properties of :class:`ModelPersister` itself.

        :return:
          A dictionary of key and value pairs, where both keys and
          values are of type ``str``.  Properties will usually include
          ``active-model`` and ``db-version`` entries.
        """

    @abstractmethod
    def upgrade(self, from_version=None, to_version=__version__):
        """Upgrade the underlying database to the latest version.

        Newer versions of Palladium may require changes to the
        :class:`ModelPersister`'s database.  This method provides an
        opportunity to run the necessary upgrade steps.

        It's the :class:`ModelPersister`'s responsibility to keep
        track of the Palladium version that was used to create and
        upgrade its database, and thus to determine the upgrade steps
        necessary.
        """


class PredictService(metaclass=ABCMeta):
    """Responsible for producing the output for the '/predict' HTTP
    endpoint.
    """
    @abstractmethod
    def __call__(self, model, request):
        """
        Use the model to run a prediction with the requested data.

        :param model:
          The :class:`~Model` instance to use for making predictions.

        :param request:
          A werkzeug ``request`` object.  A dictionary with query
          parameters is available at *request.values*.

        :return:
          A werkzeug ``response`` object.  It is the
          :class:`PredictService`'s responsiblity to return
          appropriate status codes and data in case of error.
        """



from unittest.mock import MagicMock

import pytest


class TestAnnotate:
    @pytest.fixture
    def annotate(self):
        from palladium.interfaces import annotate
        return annotate

    def test_with_existing_data(self, annotate):
        model = MagicMock(__metadata__={'one': 1, 'two': 2})
        annotate(model, {'one': '11'})
        assert model.__metadata__['one'] == '11'
        assert model.__metadata__['two'] == 2
        assert annotate(model) == {'one': '11', 'two': 2}

    def test_without_existing_data(self, annotate):
        model = TestAnnotate()
        assert annotate(model, {'one': '11'}) == {'one': '11'}
        assert model.__metadata__['one'] == '11'


def load_data_decorator(func):
    def inner(self):
        X, y = func(self)
        return self.name + X, self.name + y
    return inner


class TestDatasetLoader:
    @pytest.fixture
    def DatasetLoader(self):
        from palladium.interfaces import DatasetLoader
        return DatasetLoader

    def test_call_decorator(self, DatasetLoader, config):
        config['load_data_decorators'] = [
            'palladium.tests.test_interfaces.load_data_decorator'
            ]

        class MyDatasetLoader(DatasetLoader):
            name = 'hey'

            def __call__(self):
                return 'X', 'y'

        assert MyDatasetLoader()() == ('heyX', 'heyy')


class TestPredictError:
    @pytest.fixture
    def PredictError(self):
        from palladium.interfaces import PredictError
        return PredictError

    def test_str(self, PredictError):
        assert str(PredictError("message", 123)) == "message (123)"



import os
import readline  # https://github.com/ContinuumIO/anaconda-issues/issues/152
from unittest.mock import MagicMock
from unittest.mock import Mock

import numpy
from pandas import DataFrame
from pandas import Series
from sklearn.pipeline import Pipeline

import pytest


pytest.importorskip("rpy2")
from rpy2.robjects.pandas2ri import py2ri
from rpy2.robjects.pandas2ri import ri2py
from rpy2.robjects.vectors import Vector


@pytest.fixture
def ObjectMixin(monkeypatch):
    from palladium.R import ObjectMixin
    r_dict = {}
    r = MagicMock()
    r.__getitem__.side_effect = r_dict.__getitem__
    r.__setitem__.side_effect = r_dict.__setitem__
    r['myfunc'] = Mock()
    r['predict'] = Mock()
    monkeypatch.setattr(ObjectMixin, 'r', r)
    return ObjectMixin


class TestDatasetLoader:
    @pytest.fixture
    def DatasetLoader(self, ObjectMixin):
        from palladium.R import DatasetLoader
        return DatasetLoader

    def test_it(self, DatasetLoader):
        X, y = object(), object()
        DatasetLoader.r['myfunc'].return_value = X, y
        dloader = DatasetLoader('myscript', 'myfunc', some='kwarg')
        assert dloader() == (X, y)
        dloader.r.source.assert_called_with('myscript')
        dloader.r['myfunc'].assert_called_with(some='kwarg')


class TestAbstractModel:
    @pytest.fixture
    def data(self):
        X = numpy.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
        y = numpy.array([1, 2])
        return X, y

    @pytest.fixture
    def dataframe(self, data):
        X, y = data
        return DataFrame(X, columns=['one', 'two', 'three']), Series(y)

    @pytest.fixture
    def Model(self, ObjectMixin, monkeypatch):
        from palladium.R import AbstractModel
        monkeypatch.setattr(AbstractModel, '__abstractmethods__', set())
        return AbstractModel

    def test_fit_with_numpy_data(self, Model, data):
        X, y = data
        model = Model(scriptname='myscript', funcname='myfunc', some='kwarg')
        model.fit(X, y)
        funcargs = model.r['myfunc'].call_args
        assert (numpy.asarray(funcargs[0][0]) == X).all()
        assert (numpy.asarray(funcargs[0][1]) == y).all()
        assert funcargs[1]['some'] == 'kwarg'

    def test_fit_with_pandas_data(self, Model, dataframe):
        X, y = dataframe
        model = Model(scriptname='myscript', funcname='myfunc', some='kwarg')
        model.fit(X, y)
        funcargs = model.r['myfunc'].call_args
        assert (ri2py(funcargs[0][0]).values == X.values).all()
        assert (ri2py(funcargs[0][1]) == y).all()
        assert funcargs[1]['some'] == 'kwarg'


class TestClassificationModel(TestAbstractModel):
    @pytest.fixture
    def Model(self, ObjectMixin, monkeypatch):
        from palladium.R import ClassificationModel
        return ClassificationModel

    def test_predict_with_numpy_data(self, Model, data):
        X, y = data
        model = Model(scriptname='myscript', funcname='myfunc', some='kwarg')
        model.r['predict'].return_value = numpy.array(
            [[0.1, 0.2, 0.7], [0.8, 0.1, 0.1]])
        model.fit(X, y)

        result = model.predict(X)
        predictargs = model.r['predict'].call_args
        assert predictargs[0][0] is model.rmodel_
        assert (numpy.asarray(predictargs[0][1]) == X).all()
        assert predictargs[1]['type'] == 'prob'
        assert (result ==
                numpy.argmax(model.r['predict'].return_value, axis=1)).all()

        result = model.predict_proba(X)
        assert (result == model.r['predict'].return_value).all()

    def test_predict_with_pandas_data(self, Model, dataframe):
        X, y = dataframe
        model = Model(scriptname='myscript', funcname='myfunc', some='kwarg')
        model.r['predict'].return_value = numpy.array(
            [[0.1, 0.2, 0.7], [0.8, 0.1, 0.1]])
        model.fit(X, y)

        result = model.predict(X)
        predictargs = model.r['predict'].call_args
        assert predictargs[0][0] is model.rmodel_
        assert (ri2py(predictargs[0][1]).values == X.values).all()
        assert predictargs[1]['type'] == 'prob'
        assert (result ==
                numpy.argmax(model.r['predict'].return_value, axis=1)).all()

        result = model.predict_proba(X)
        assert (result == model.r['predict'].return_value).all()


class TestClassification:
    @pytest.fixture
    def dataset(self):
        from palladium.R import DatasetLoader
        return DatasetLoader(
            scriptname=os.path.join(os.path.dirname(__file__), 'test_R.R'),
            funcname='dataset',
            )

    @pytest.fixture
    def model(self):
        from palladium.R import ClassificationModel
        return ClassificationModel(
            scriptname=os.path.join(os.path.dirname(__file__), 'test_R.R'),
            funcname='train.randomForest',
            encode_labels=True,
            )

    def test_fit_and_predict(self, dataset, model):
        X, y = dataset()
        model.fit(X, y)
        probas = model.predict_proba(X)
        assert probas.shape == (150, 3)
        assert (probas.sum(axis=1) == 1).all()
        assert (model.predict(X) == numpy.asarray(y)).all()

    def test_fit_and_score(self, dataset, model):
        X, y = dataset()
        model.fit(X, y)
        assert model.score(X, y) == 1.0


class TestClassificationWithNumpyDataset(TestClassification):
    @pytest.fixture
    def dataset(self):
        X = numpy.array([
            [1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0],
            [7.0, 8.0, 9.0],
            ] * 50)
        y = numpy.array([1, 2, 3] * 50)
        return lambda: (X, y)


class TestClassificationWithPandasDataset(TestClassification):
    @pytest.fixture
    def dataset(self):
        X = DataFrame([
            [1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0],
            [7.0, 8.0, 9.0],
            ] * 50)
        y = Series([1, 2, 3] * 50)
        return lambda: (X, y)


class TestRegression:
    @pytest.fixture
    def scriptname(self):
        return os.path.join(os.path.dirname(__file__), 'test_R2.R')

    @pytest.fixture
    def dataset(self, scriptname):
        from palladium.R import DatasetLoader
        return DatasetLoader(
            scriptname=scriptname,
            funcname='dataset',
            )

    @pytest.fixture
    def model(self, scriptname):
        from palladium.R import RegressionModel
        from palladium.R import Rpy2Transform

        return Pipeline([
            ('rpy2', Rpy2Transform()),
            ('regressor', RegressionModel(
                scriptname=scriptname,
                funcname='train.randomForest',
                )),
            ])

    def test_smoke(self, dataset, model):
        X, y = dataset()
        model.fit(X, y)

        assert model.predict(X).shape == (
            ri2py(y).shape if isinstance(y, Vector) else y.shape)
        score_1 = model.score(X, y)
        assert score_1 >= 0.1

        # Convert X to its Python or R equivalent and check if scores
        # match:
        X_t = py2ri(X) if isinstance(X, DataFrame) else ri2py(X)
        score_2 = model.score(X_t, y)
        assert score_2 == score_1

        # Convert X to a Python list and run the prediction:
        X_t2 = ri2py(X) if not isinstance(X, DataFrame) else X
        X_t2 = X_t2.values.tolist()
        score_3 = model.score(X_t2, y)
        assert score_3 == score_1


class TestRegressionWithPandasDataset(TestRegression):
    @pytest.fixture
    def dataset(self):
        X = DataFrame(
            [
                [1.0, 2.0, 0],
                [4.0, 5.0, 1],
                [7.0, 8.0, 2],
            ] * 50,
            columns=('one', 'two', 'really'),
            )
        X['really'] = X['really'].astype('category')
        y = Series([1.0, 2.0, 3.0] * 50)
        return lambda: (X, y)

    def test_fit_and_predict(self, dataset, model):
        X, y = dataset()
        model.fit(X, y)
        assert (model.predict(X) == numpy.asarray(y)).all()

    def test_fit_and_score(self, dataset, model):
        X, y = dataset()
        model.fit(X, y)
        assert model.score(X, y) == 1.0



from unittest.mock import call
from unittest.mock import Mock
from unittest.mock import patch

import numpy as np
import pytest
from pytest import fixture


pytest.importorskip("julia")


@fixture
def bridge(monkeypatch):
    make_bridge = Mock()
    monkeypatch.setattr('palladium.julia.make_bridge', make_bridge)
    return make_bridge.return_value


class TestAbstractModel:
    @fixture
    def Model(self):
        from palladium.julia import AbstractModel
        return AbstractModel

    @fixture
    def model(self, Model, bridge):
        return Model(
            fit_func='myjulia.fit_func',
            predict_func='yourjulia.predict_func',
            fit_kwargs={'fit': 'kwargs'},
            predict_kwargs={'predict': 'kwargs'},
            )

    def test_initialize_julia(self, model, bridge):
        bridge.eval.side_effect = ['fit', 'predict']
        model._initialize_julia()
        assert model.fit_func_ == 'fit'
        assert model.predict_func_ == 'predict'

        assert call('import myjulia') in bridge.mock_calls
        assert call('import yourjulia') in bridge.mock_calls
        assert call('myjulia.fit_func') in bridge.mock_calls
        assert call('yourjulia.predict_func') in bridge.mock_calls

    def test_fit(self, model):
        X, y = Mock(), Mock()
        assert model.fit(X, y) is model
        fit_func = model.fit_func_
        fit_func.assert_called_with(X.T, y, fit='kwargs')
        assert model.fitted_ == fit_func.return_value

    def test_fit_with_label_encoder(self, model):
        model.encode_labels = True
        X, y = Mock(), Mock()
        with patch('palladium.julia.LabelEncoder') as encoder:
            model.fit(X, y) is model
        fit_func = model.fit_func_
        transform = encoder().fit_transform
        fit_func.assert_called_with(X.T, transform.return_value, fit='kwargs')

    def test_predict(self, model):
        X = Mock()
        model.fitted_ = Mock()
        model._initialize_julia()
        result = model.predict(X)
        predict_func = model.predict_func_
        predict_func.assert_called_with(
            model.fitted_, X.astype().T, predict='kwargs')
        assert result == predict_func.return_value

    def test_predict_with_label_encoder(self, model):
        model.encode_labels = True
        X = Mock()
        model.fitted_ = Mock()
        model.enc_ = Mock()
        inverse_transform = model.enc_.inverse_transform
        model._initialize_julia()
        result = model.predict(X)
        assert result == inverse_transform.return_value
        inverse_transform.assert_called_with(model.predict_func_.return_value)

    def test_predict_convert_to_float(self, model):
        X_in = np.array([1, 2, 3], dtype=object)
        model.fitted_ = Mock()
        model.enc_ = Mock()
        model._initialize_julia()
        model.predict(X_in)
        X_conv = model.predict_func_.call_args[0][1]
        assert np.all(X_in == X_conv)
        assert X_conv.dtype == np.float64

    def test_getstate(self, model):
        model._initialize_julia()
        model.fitted_ = Mock()

        state = model.__getstate__()
        assert state['fitted_'] is not model.fitted_
        assert 'fit_func_' not in state
        assert 'predict_func_' not in state
        assert 'bridge_' not in state

        idict = model.__dict__
        assert 'fit_func_' in idict
        assert 'predict_func_' in idict
        assert 'bridge_' in idict

    def test_setstate(self, model, bridge):
        model.__setstate__(
            {'fitted_': 'fitted', 'fit_kwargs': {'bl': 'arg'}})
        assert model.fitted_ != 'fitted'
        assert model.fitted_ == bridge.eval.return_value.return_value


class TestClassificationModel(TestAbstractModel):
    @fixture
    def Model(self):
        from palladium.julia import ClassificationModel
        return ClassificationModel

    def test_score(self, model):
        X, y = Mock(), Mock()
        with patch('palladium.julia.accuracy_score') as accuracy_score:
            with patch('palladium.julia.AbstractModel.predict') as predict:
                model.score(X, y)
        accuracy_score.assert_called_with(predict.return_value, y)
        predict.assert_called_with(X)



"""Utilities for testing the performance of a trained model.
"""

from pprint import pprint
import sys

from docopt import docopt
from sklearn.metrics import get_scorer

from .util import args_from_config
from .util import initialize_config
from .util import logger
from .util import timer


@args_from_config
def test(dataset_loader_test, model_persister,
         scoring=None, model_version=None):

    with timer(logger.info, "Loading data"):
        X, y = dataset_loader_test()

    with timer(logger.info, "Reading model"):
        model = model_persister.read(version=model_version)

    logger.info(
        'Loaded model version {}'.format(model.__metadata__['version']))

    if not (hasattr(model, 'score') or scoring is not None):
        raise ValueError(
            "Your model doesn't seem to implement a 'score' method.  You may "
            "want to define a 'scoring' option in the configuration."
            )

    with timer(logger.info, "Applying model"):
        scores = []
        if scoring is not None:
            if not isinstance(scoring, dict):
                scoring = {'score': scoring}
            for key, scorer in scoring.items():
                scorer = get_scorer(scorer)
                scores.append("{}: {}".format(key, scorer(model, X, y)))
        else:
            scores.append("score: {}".format(model.score(X, y)))

    logger.info("Score: {}.".format('\n       '.join(scores)))
    return scores


def test_cmd(argv=sys.argv[1:]):  # pragma: no cover
    """\
Test a model.

Uses 'dataset_loader_test' and 'model_persister' from the
configuration to load a test dataset to test the accuracy of a trained
model with.

Usage:
  pld-test [options]

Options:
  -h --help                  Show this screen.

  --model-version=<version>  The version of the model to be tested. If
                             not specified, the newest model will be used.
"""
    arguments = docopt(test_cmd.__doc__, argv=argv)
    model_version = arguments['--model-version']
    model_version = int(model_version) if model_version is not None else None
    initialize_config(__mode__='fit')
    test(model_version=model_version)


@args_from_config
def list(model_persister):
    print("Models:")
    pprint(model_persister.list_models())
    print("Properties:")
    pprint(model_persister.list_properties())


def list_cmd(argv=sys.argv[1:]):  # pragma: no cover
    """\
List information about available models.

Uses the 'model_persister' from the configuration to display a list of
models and their metadata.

Usage:
  pld-list [options]

Options:
  -h --help                  Show this screen.
"""
    docopt(list_cmd.__doc__, argv=argv)
    initialize_config(__mode__='fit')
    list()



"""Utilities for fitting modles.
"""

import gc
from warnings import warn
import sys

from datetime import datetime
from docopt import docopt
from joblib import parallel_backend
import pandas
from sklearn.metrics import get_scorer
from sklearn.model_selection import GridSearchCV

from .interfaces import annotate
from .util import apply_kwargs
from .util import args_from_config
from .util import initialize_config
from .util import logger
from .util import PluggableDecorator
from .util import timer


def _persist_model(model, model_persister, activate=True):
    metadata = {
        'train_timestamp': datetime.now().isoformat(),
        }
    cv_results = getattr(model, 'cv_results_', None)
    if cv_results is not None:
        json_str = pandas.DataFrame(cv_results).to_json(orient='records')
        metadata['cv_results'] = json_str
    annotate(model, metadata)
    with timer(logger.info, "Writing model"):
        version = model_persister.write(model)
    logger.info("Wrote model with version {}.".format(version))
    if activate:
        model_persister.activate(version)


@PluggableDecorator('fit_decorators')
@args_from_config
def fit(dataset_loader_train, model, model_persister, persist=True,
        activate=True, dataset_loader_test=None, evaluate=False,
        persist_if_better_than=None, scoring=None):

    if persist_if_better_than is not None:
        evaluate = True
        if dataset_loader_test is None:
            raise ValueError(
                "When using 'persist_if_better_than', make sure you also "
                "provide a 'dataset_loader_test'."
                )

    if evaluate and not (hasattr(model, 'score') or scoring is not None):
        raise ValueError(
            "Your model doesn't seem to implement a 'score' method.  You may "
            "want to define a 'scoring' option in the configuration."
            )

    if scoring is not None:
        scorer = get_scorer(scoring)
    else:
        def scorer(model, X, y):
            return model.score(X, y)

    with timer(logger.info, "Loading data"):
        X, y = dataset_loader_train()

    with timer(logger.info, "Fitting model"):
        model.fit(X, y)

    if evaluate:
        with timer(logger.debug, "Evaluating model on train set"):
            score_train = scorer(model, X, y)
            annotate(model, {'score_train': score_train})
            logger.info("Train score: {}".format(score_train))

    X, y = None, None
    gc.collect()

    score_test = None
    if evaluate and dataset_loader_test is not None:
        with timer(logger.info, "Loading test data"):
            X_test, y_test = dataset_loader_test()
        with timer(logger.debug, "Evaluating model on test set"):
            score_test = scorer(model, X_test, y_test)
            annotate(model, {'score_test': score_test})
            logger.info("Test score:  {}".format(score_test))

    if persist:
        if (persist_if_better_than is not None and
            score_test < persist_if_better_than):
            logger.info("Not persisting model that has a test score "
                        "{} < {}".format(score_test, persist_if_better_than))
        else:
            _persist_model(model, model_persister, activate=activate)

    return model


def fit_cmd(argv=sys.argv[1:]):  # pragma: no cover
    """\
Fit a model and save to database.

Will use 'dataset_loader_train', 'model', and 'model_perister' from
the configuration file, to load a dataset to train a model with, and
persist it.

Usage:
  pld-fit [options]

Options:
  -n --no-save              Don't persist the fitted model to disk.

  --no-activate             Don't activate the fitted model.

  --save-if-better-than=<k> Persist only if test score better than given
                            value.

  -e --evaluate             Evaluate fitted model on train and test set and
                            print out results.

  -h --help                 Show this screen.
"""
    arguments = docopt(fit_cmd.__doc__, argv=argv)
    no_save = arguments['--no-save']
    no_activate = arguments['--no-activate']
    save_if_better_than = arguments['--save-if-better-than']
    evaluate = arguments['--evaluate'] or bool(save_if_better_than)
    if save_if_better_than is not None:
        save_if_better_than = float(save_if_better_than)
    initialize_config(__mode__='fit')
    fit(
        persist=not no_save,
        activate=not no_activate,
        evaluate=evaluate,
        persist_if_better_than=save_if_better_than,
        )


@args_from_config
def activate(model_persister, model_version):
    model_persister.activate(model_version)
    logger.info("Activated model with version {}.".format(model_version))


@args_from_config
def delete(model_persister, model_version):
    model_persister.delete(model_version)
    logger.info("Deleted model with version {}.".format(model_version))


def admin_cmd(argv=sys.argv[1:]):  # pragma: no cover
    """\
Activate or delete models.

Models are usually made active right after fitting (see command
pld-fit).  The 'activate' command allows you to explicitly set the
currently active model.  Use 'pld-list' to get an overview of all
available models along with their version identifiers.

Deleting a model will simply remove it from the database.

Usage:
  pld-admin activate <version> [options]
  pld-admin delete <version> [options]

Options:
  -h --help                 Show this screen.
"""
    arguments = docopt(admin_cmd.__doc__, argv=argv)
    initialize_config(__mode__='fit')
    if arguments['activate']:
        activate(model_version=int(arguments['<version>']))
    elif arguments['delete']:
        delete(model_version=int(arguments['<version>']))


def with_parallel_backend(
    estimator,
    backend,
    methods=('fit', 'predict', 'predict_proba'),
    **backend_params
):
    def wrapper(func):
        def wrapped(*args, **kwargs):
            with parallel_backend(backend, **backend_params):
                return func(*args, **kwargs)
        return wrapped

    for name in methods:
        setattr(estimator, name, wrapper(getattr(estimator, name)))
    return estimator


@args_from_config
def grid_search(dataset_loader_train, model, grid_search, scoring=None,
                save_results=None, persist_best=False, model_persister=None):
    if persist_best and model_persister is None:
        raise ValueError(
            "Cannot persist the best model without a model_persister. Please "
            "specify one in your Palladium configuration file."
            )

    with timer(logger.info, "Loading data"):
        X, y = dataset_loader_train()

    if isinstance(grid_search, dict):
        search_kwargs = {
            'refit': persist_best,
            }
        search_kwargs.update(grid_search)

        cv = search_kwargs.get('cv', None)
        if callable(cv):
            search_kwargs['cv'] = apply_kwargs(cv, n=len(y), X=X, y=y)

        if 'scoring' in search_kwargs:
            if scoring is not None:
                raise ValueError("You cannot define 'scoring' in 'grid_search' "
                                 "and globally.")
            warn("Use of 'scoring' inside of 'grid_search' is deprecated. "
                 "To fix, move 'scoring' up to the top level of the configuration "
                 "dict.", DeprecationWarning)
            scoring = search_kwargs['scoring']
        elif scoring is not None:
            search_kwargs['scoring'] = scoring

        if not (hasattr(model, 'score') or scoring is not None):
            raise ValueError(
                "Your model doesn't seem to implement a 'score' method.  You may "
                "want to define a 'scoring' option in the configuration."
                )

        search = GridSearchCV(model, **search_kwargs)
    else:
        search = grid_search

    with timer(logger.info, "Running grid search"):
        search.fit(X, y)

    results = pandas.DataFrame(search.cv_results_)
    pandas.options.display.max_rows = len(results)
    pandas.options.display.max_columns = len(results.columns)
    if 'rank_test_score' in results:
        results = results.sort_values('rank_test_score')
    print(results)
    if save_results:
        results.to_csv(save_results, index=False)
    if persist_best:
        _persist_model(search, model_persister, activate=True)
    return search


def grid_search_cmd(argv=sys.argv[1:]):  # pragma: no cover
    """\
Grid search parameters for the model.

Uses 'dataset_loader_train', 'model', and 'grid_search' from the
configuration to load a training dataset, and run a grid search on the
model using the grid of hyperparameters.

Usage:
  pld-grid-search [options]

Options:
  --save-results=<fname>   Save results to CSV file
  --persist-best           Persist the best model from grid search
  -h --help                Show this screen.
"""
    arguments = docopt(grid_search_cmd.__doc__, argv=argv)
    initialize_config(__mode__='fit')
    grid_search(
        save_results=arguments['--save-results'],
        persist_best=arguments['--persist-best'],
        )



from datetime import datetime
from functools import partial
import numpy as np
from unittest.mock import Mock
from unittest.mock import call
from unittest.mock import patch

from dateutil.parser import parse
import pandas
import pytest


class TestFit:
    @pytest.fixture
    def fit(self):
        from palladium.fit import fit
        return fit

    @pytest.fixture
    def dataset_loader(self):
        dataset_loader = Mock()
        dataset_loader.return_value = Mock(), Mock()
        return dataset_loader

    def test_it(self, fit):
        model, dataset_loader_train, model_persister = Mock(), Mock(), Mock()
        del model.cv_results_
        X, y = object(), object()
        dataset_loader_train.return_value = X, y

        result = fit(dataset_loader_train, model, model_persister)

        assert result is model
        dataset_loader_train.assert_called_with()
        model.fit.assert_called_with(X, y)
        model_persister.write.assert_called_with(model)
        model_persister.activate.assert_called_with(
            model_persister.write.return_value)

    def test_no_persist(self, fit):
        model, dataset_loader_train, model_persister = Mock(), Mock(), Mock()
        del model.cv_results_
        X, y = object(), object()
        dataset_loader_train.return_value = X, y

        result = fit(dataset_loader_train, model, model_persister,
                     persist=False)

        assert result is model
        dataset_loader_train.assert_called_with()
        model.fit.assert_called_with(X, y)
        assert model_persister.call_count == 0

    def test_evaluate_no_test_dataset(self, fit):
        model, dataset_loader_train, model_persister = Mock(), Mock(), Mock()
        del model.cv_results_
        X, y = object(), object()
        dataset_loader_train.return_value = X, y

        result = fit(dataset_loader_train, model, model_persister,
                     evaluate=True)

        assert result is model
        dataset_loader_train.assert_called_with()
        model.fit.assert_called_with(X, y)
        assert model.score.call_count == 1
        model.score.assert_called_with(X, y)
        model_persister.write.assert_called_with(model)

    def test_evaluate_with_test_dataset(self, fit):
        model, dataset_loader_train, model_persister = Mock(), Mock(), Mock()
        del model.cv_results_
        dataset_loader_test = Mock()
        X, y, X_test, y_test = object(), object(), object(), object()
        dataset_loader_train.return_value = X, y
        dataset_loader_test.return_value = X_test, y_test

        result = fit(dataset_loader_train, model, model_persister,
                     dataset_loader_test=dataset_loader_test,
                     evaluate=True)

        assert result is model
        dataset_loader_train.assert_called_with()
        dataset_loader_test.assert_called_with()
        model.fit.assert_called_with(X, y)
        assert model.score.call_count == 2
        assert model.score.mock_calls[0] == call(X, y)
        assert model.score.mock_calls[1] == call(X_test, y_test)
        model_persister.write.assert_called_with(model)

    def test_evaluate_annotations(self, fit, dataset_loader):
        model = Mock()
        del model.cv_results_
        model.score.side_effect = [0.9, 0.8]

        result = fit(
            dataset_loader_train=dataset_loader,
            model=model,
            model_persister=Mock(),
            dataset_loader_test=dataset_loader,
            persist_if_better_than=0.9,
            )

        assert result.__metadata__['score_train'] == 0.9
        assert result.__metadata__['score_test'] == 0.8

    def test_evaluate_scoring(self, fit, dataset_loader):
        model = Mock()
        del model.cv_results_
        scorer = Mock()
        scorer.side_effect = [0.99, 0.01]

        fit(
            dataset_loader_train=dataset_loader,
            model=model,
            model_persister=Mock(),
            dataset_loader_test=dataset_loader,
            scoring=scorer,
            evaluate=True,
            )
        assert model.score.call_count == 0
        assert scorer.call_count == 2

    def test_evaluate_no_score(self, fit, dataset_loader):
        model = Mock()
        del model.score
        del model.cv_results_

        with pytest.raises(ValueError):
            fit(
                dataset_loader_train=dataset_loader,
                model=model,
                model_persister=Mock(),
                dataset_loader_test=dataset_loader,
                evaluate=True,
                )

    def test_persist_if_better_than(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        model.score.return_value = 0.9
        del model.cv_results_

        result = fit(
            dataset_loader_train=dataset_loader,
            model=model,
            model_persister=model_persister,
            dataset_loader_test=dataset_loader,
            persist_if_better_than=0.9,
            )

        assert result is model
        assert model_persister.write.call_count == 1

    def test_persist_if_better_than_false(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        model.score.return_value = 0.9
        del model.cv_results_

        result = fit(
            dataset_loader_train=dataset_loader,
            model=model,
            model_persister=model_persister,
            dataset_loader_test=dataset_loader,
            persist_if_better_than=0.91,
            )

        assert result is model
        assert model_persister.write.call_count == 0

    def test_persist_if_better_than_persist_false(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        model.score.return_value = 0.9
        del model.cv_results_

        result = fit(
            dataset_loader_train=dataset_loader,
            model=model,
            model_persister=model_persister,
            persist=False,
            dataset_loader_test=dataset_loader,
            persist_if_better_than=0.9,
            )

        assert result is model
        assert model_persister.write.call_count == 0

    def test_persist_if_better_than_no_dataset_test(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        model.score.return_value = 0.9
        del model.cv_results_

        with pytest.raises(ValueError):
            fit(
                dataset_loader_train=dataset_loader,
                model=model,
                model_persister=model_persister,
                dataset_loader_test=None,
                persist_if_better_than=0.9,
                )

    def test_activate_no_persist(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        del model.cv_results_

        result = fit(
            dataset_loader_train=dataset_loader,
            model=model,
            model_persister=model_persister,
            persist=False,
            )
        assert result is model
        assert model_persister.activate.call_count == 0

    def test_timestamp(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        del model.cv_results_

        def persist(model):
            assert 'train_timestamp' in model.__metadata__
        model_persister.write.side_effect = persist

        before_fit = datetime.now()
        result = fit(
            dataset_loader,
            model,
            model_persister,
            )
        after_fit = datetime.now()

        assert result is model

        timestamp = parse(model.__metadata__['train_timestamp'])
        assert before_fit < timestamp < after_fit
        model_persister.write.assert_called_with(model)

    def test_cv_results(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        model.cv_results_ = {
            'mean_train_score': [3, 2, 1],
            'mean_test_score': [1, 2, 3],
            }

        def persist(model):
            assert 'cv_results' in model.__metadata__

        model_persister.write.side_effect = persist

        result = fit(
            dataset_loader,
            model,
            model_persister,
            )
        assert result is model

        cv_results = model.__metadata__['cv_results']
        cv_results = pandas.read_json(cv_results).to_dict(orient='list')
        assert cv_results == model.cv_results_
        model_persister.write.assert_called_with(model)


def test_activate():
    from palladium.fit import activate

    persister = Mock()
    activate(persister, 2)
    persister.activate.assert_called_with(2)


def test_delete():
    from palladium.fit import delete

    persister = Mock()
    delete(persister, 2)
    persister.delete.assert_called_with(2)


class TestGridSearch:
    @pytest.fixture
    def GridSearchCVWithScores(self, monkeypatch):
        scores = {
            'mean_test_score': [0.1, 0.2],
            'std_test_score': [0.06463643, 0.05073433],
            'params': [{'C': 0.1}, {'C': 0.3}],
            'rank_test_score': [1, 2],
            }

        GridSearchCV = Mock()
        monkeypatch.setattr(
            'palladium.fit.GridSearchCV', GridSearchCV)
        GridSearchCV().cv_results_ = scores
        return GridSearchCV

    @pytest.fixture
    def grid_search(self):
        from palladium.fit import grid_search
        return grid_search

    def test_it(self, grid_search, GridSearchCVWithScores, capsys, tmpdir):
        model, dataset_loader_train = Mock(), Mock()
        grid_search_params = {'verbose': 4}
        X, y = object(), object()
        dataset_loader_train.return_value = X, y

        results_csv = tmpdir.join('results.csv')
        result = grid_search(
            dataset_loader_train=dataset_loader_train,
            model=model,
            grid_search=grid_search_params,
            save_results=str(results_csv),
            )
        dataset_loader_train.assert_called_with()
        GridSearchCVWithScores.assert_called_with(model, refit=False, verbose=4)
        GridSearchCVWithScores().fit.assert_called_with(X, y)
        assert result is GridSearchCVWithScores()
        scores = GridSearchCVWithScores().cv_results_
        assert (str(pandas.DataFrame(scores)).strip() ==
                capsys.readouterr()[0].strip())
        assert (str(pandas.DataFrame(scores)).strip() ==
                str(pandas.read_csv(str(results_csv))).strip())

    def test_no_score_method_raises(self, grid_search):
        model, dataset_loader_train = Mock(spec=['fit', 'predict']), Mock()
        dataset_loader_train.return_value = object(), object()

        with pytest.raises(ValueError):
            grid_search(dataset_loader_train, model, {})

    def test_two_scores_raises(self, grid_search):
        model, dataset_loader_train = Mock(spec=['fit', 'predict']), Mock()
        dataset_loader_train.return_value = object(), object()

        with pytest.raises(ValueError):
            grid_search(dataset_loader_train, model,
                        {'scoring': 'f1'}, scoring='accuracy')

    def test_two_scores_priority(self, grid_search, GridSearchCVWithScores):
        # 'scoring' has higher priority than 'model.score'
        model = Mock(spec=['fit', 'predict', 'score'])
        dataset_loader_train = Mock()
        scoring = Mock()
        dataset_loader_train.return_value = object(), object()

        grid_search(dataset_loader_train, model, {}, scoring=scoring)
        GridSearchCVWithScores.assert_called_with(
            model, refit=False, scoring=scoring)

    def test_deprecated_scoring(self, grid_search, GridSearchCVWithScores):
        # 'scoring' inside of 'grid_search' is deprecated
        model = Mock(spec=['fit', 'predict', 'score'])
        dataset_loader_train = Mock()
        scoring = Mock()
        dataset_loader_train.return_value = object(), object()

        with pytest.warns(DeprecationWarning):
            grid_search(dataset_loader_train, model,
                        {'scoring': scoring}, scoring=None)
        GridSearchCVWithScores.assert_called_with(
            model, refit=False, scoring=scoring)

    def test_persist_best_requires_persister(self, grid_search):
        model = Mock(spec=['fit', 'predict'])
        del model.cv_results_
        dataset_loader_train = Mock()
        scoring = Mock()
        dataset_loader_train.return_value = object(), object()

        with pytest.raises(ValueError):
            grid_search(dataset_loader_train, model, {}, scoring=scoring,
                        persist_best=True)

    def test_persist_best(self, grid_search, GridSearchCVWithScores):
        model = Mock(spec=['fit', 'predict'])
        del model.cv_results_
        dataset_loader_train = Mock()
        scoring = Mock()
        model_persister = Mock()
        dataset_loader_train.return_value = object(), object()

        grid_search(dataset_loader_train, model, {}, scoring=scoring,
                    persist_best=True, model_persister=model_persister)
        GridSearchCVWithScores.assert_called_with(
            model, refit=True, scoring=scoring)
        model_persister.write.assert_called_with(
            GridSearchCVWithScores())

    def test_grid_search(self, grid_search):
        model, dataset_loader_train = Mock(), Mock()
        dataset_loader_train.return_value = (
            np.random.random((10, 10)), np.random.random(10))

        CVIterator = Mock()

        def cv_iterator(n, p):
            return CVIterator(n=n, p=p)

        grid_search_params = {'cv': partial(cv_iterator, p=2)}

        scores = {
            'mean_test_score': [0.1, 0.2],
            'std_test_score': [0.06463643, 0.05073433],
            'params': [{'C': 0.1}, {'C': 0.3}],
            }
        with patch('palladium.fit.GridSearchCV') as GridSearchCV:
            GridSearchCV().cv_results_ = scores
            grid_search(dataset_loader_train, model, grid_search_params)

        GridSearchCV.assert_called_with(model, refit=False,
                                        cv=CVIterator.return_value)
        CVIterator.assert_called_with(n=10, p=2)

    def test_grid_search_with_instance(self, grid_search):
        scores = {
            'mean_test_score': [0.1, 0.2],
            'std_test_score': [0.06463643, 0.05073433],
            'params': [{'C': 0.1}, {'C': 0.3}],
            }
        model, dataset_loader_train = Mock(), Mock()
        X, y = np.random.random((10, 10)), np.random.random(10)
        dataset_loader_train.return_value = X, y
        search = Mock()
        search.cv_results_ = scores
        grid_search(dataset_loader_train, model, grid_search=search)
        search.fit.assert_called_with(X, y)


class TestWithParallelBackend:
    @pytest.fixture
    def with_parallel_backend(self):
        from palladium.fit import with_parallel_backend
        return with_parallel_backend

    @pytest.fixture
    def estimator(self):
        from sklearn.model_selection import GridSearchCV
        from sklearn.linear_model import LogisticRegression

        return GridSearchCV(
            LogisticRegression(solver='liblinear'),
            param_grid={'C': [0.001, 0.01]},
            cv=3,
            )

    @pytest.mark.parametrize('backend', ['threading', 'sequential'])
    def test_it(self, with_parallel_backend, estimator, backend):
        X, y = np.random.random((100, 10)), np.random.randint(0, 2, 100)
        with_parallel_backend(estimator, backend).fit(X, y)
        with_parallel_backend(estimator, backend).predict(X)

    def test_bad(self, with_parallel_backend, estimator):
        X, y = np.random.random((100, 10)), np.random.randint(0, 2, 100)
        with pytest.raises(KeyError):
            with_parallel_backend(estimator, 'foo').fit(X, y)


class TestFitMode():
    # test if fit mode is set in non-server scripts

    @pytest.mark.parametrize("func, cmd, argv", [
        ('palladium.fit.fit', 'palladium.fit.fit_cmd', ()),
        ('palladium.fit.grid_search', 'palladium.fit.grid_search_cmd', ()),
        ('palladium.fit.activate', 'palladium.fit.admin_cmd',
         ('activate', '1')),
        ('palladium.eval.test', 'palladium.eval.test_cmd', ()),
        ('palladium.eval.list', 'palladium.eval.list_cmd', ()),
        ('palladium.util.upgrade', 'palladium.util.upgrade_cmd', ()),
        ])
    def test_check_fit_mode(self, func, cmd, argv, config):
        from palladium.util import resolve_dotted_name
        with patch(func):
            cmd = resolve_dotted_name(cmd)
            assert config.initialized is False
            cmd(argv=argv)
            assert config.initialized is True
            assert config == {'__mode__': 'fit'}



"""Support for building models using the R programming language.
"""

from palladium.interfaces import DatasetLoader
from palladium.interfaces import Model
import numpy as np
from pandas import Categorical
from pandas import DataFrame
from pandas import Series
from rpy2 import robjects
from rpy2.robjects import pandas2ri
from rpy2.robjects.pandas2ri import py2ri
from rpy2.robjects.numpy2ri import numpy2ri
from sklearn.base import TransformerMixin
from sklearn.metrics import accuracy_score
from sklearn.metrics import r2_score
from sklearn.preprocessing import LabelEncoder


pandas2ri.activate()


class ObjectMixin:
    r = robjects.r

    def __init__(self, scriptname, funcname, **kwargs):
        self.scriptname = scriptname
        self.funcname = funcname
        self.r.source(scriptname)
        self.rfunc = self.r[funcname]
        self.kwargs = kwargs


class DatasetLoader(DatasetLoader, ObjectMixin):
    """A :class:`~palladium.interfaces.DatasetLoader` that calls an R
    function to load the data.
    """
    def __call__(self):
        X, y = self.rfunc(**self.kwargs)
        return X, y


class AbstractModel(Model, ObjectMixin):
    def __init__(self, encode_labels=False, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)
        self.encode_labels = encode_labels

    @staticmethod
    def _from_python(obj):
        if isinstance(obj, DataFrame):
            obj = py2ri(obj)
        elif isinstance(obj, Series):
            obj = numpy2ri(obj.values)
        elif isinstance(obj, np.ndarray):
            obj = numpy2ri(obj)
        return obj

    def fit(self, X, y=None):
        if self.encode_labels:
            self.enc_ = LabelEncoder()
            y = self.enc_.fit_transform(y)

        self.rmodel_ = self.rfunc(
            self._from_python(X),
            self._from_python(y),
            **self.kwargs)


class ClassificationModel(AbstractModel):
    """A :class:`~palladium.interfaces.Model` for classification problems
    that uses an R model for training and prediction.
    """

    def predict_proba(self, X):
        X = self._from_python(X)
        return np.asarray(self.r['predict'](self.rmodel_, X, type='prob'))

    def predict(self, X):
        X = X.astype(float) if hasattr(X, 'astype') else X
        y_pred = np.argmax(self.predict_proba(X), axis=1)
        if self.encode_labels:
            y_pred = self.enc_.inverse_transform(y_pred)
        return y_pred

    def score(self, X, y):
        return accuracy_score(self.predict(X), np.asarray(y))


class RegressionModel(AbstractModel):
    """A :class:`~palladium.interfaces.Model` for regression problems
    that uses an R model for training and prediction.
    """

    def predict(self, X):
        X = self._from_python(X)
        return np.asarray(self.r['predict'](self.rmodel_, X))

    def score(self, X, y):
        return r2_score(self.predict(X), np.asarray(y))


class Rpy2Transform(TransformerMixin):
    def fit(self, X, y):
        if isinstance(X, np.ndarray):
            pass
        elif isinstance(X, DataFrame):
            self.index2levels_ = {}
            for index, column in enumerate(X.columns):
                if hasattr(X[column].dtype, 'categories'):
                    self.index2levels_[index] = tuple(
                        X[column].dtype.categories)
            self.colnames_ = list(X.columns)
        else:
            self.index2levels_ = {}
            for index in range(len(X.colnames)):
                if hasattr(X[index], 'levels'):
                    self.index2levels_[index] = tuple(X[index].levels)
            self.colnames_ = X.colnames
        return self

    def transform(self, X):
        if isinstance(X, (np.ndarray, list)) and hasattr(self, 'index2levels_'):
            X = DataFrame(X, columns=self.colnames_)
        if isinstance(X, DataFrame) and hasattr(self, 'index2levels_'):
            for index, levels in self.index2levels_.items():
                colname = X.columns[index]
                X[colname] = Categorical(
                    X[colname],
                    categories=levels,
                    )
            X = py2ri(X)
        if hasattr(self, 'colnames_'):
            # Deal with an rpy2 issue whereas colnames appear to get
            # mangled when calling py2ri.  Also, apply colnames if
            # predict data was missing them:
            X.colnames = self.colnames_
        return X



"""The cache module provides caching utilities in order to provide
faster access to data which is needed repeatedly.  The disk cache
(:class:`diskcache`) which is primarily used during development when
loading data from the local harddisk is faster than querying a remote
database.
"""

import hashlib
from functools import wraps
import os
import pickle
from tempfile import gettempdir

from joblib import numpy_pickle


class abstractcache(object):
    """
    An abstract class for providing basic functionality for caching
    function calls. It contains the handling of keys used for
    caching objects.
    """
    cache = None  # a dict-like that has to be implemented

    def __init__(self, compute_key=None, ignore=False):
        self.compute_key = compute_key
        self.ignore = ignore

    def __call__(self, func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            if self.ignore() if callable(self.ignore) else self.ignore:
                return func(*args, **kwargs)

            if self.compute_key is not None:
                key = self.compute_key(*args, **kwargs)
            else:
                key = args, tuple(kwargs.items())
            try:
                return self.cache[key]
            except KeyError:
                pass

            value = func(*args, **kwargs)
            self.cache[key] = value
            return value

        self.func = func
        wrapped.__cache__ = self
        return wrapped

    def __len__(self):  # pragma: no cover
        raise NotImplementedError()


class diskcache(abstractcache):
    """
    The disk cache stores results of function calls as pickled files
    to disk.  Usually used during development and evaluation to save
    costly DB interactions in repeated calls with the same data.

    Note: Should changes to the database or to your functions require
    you to purge existing cached values, then those cache files are
    found in the location defined in :attr:`filename_tmpl`.
    """
    def __init__(self, *args, filename_tmpl=None, **kwargs):
        """
        :param str filename_tmpl:
          The filename template that I will use to store cache files,
          e.g. ``{}``.
        """.format(diskcache.filename_tmpl)
        super().__init__(*args, **kwargs)
        if filename_tmpl is not None:
            if '{key}' not in filename_tmpl:
                raise ValueError(
                    "Your filename_tmpl must have a {key} placeholder,"
                    "e.g., cache-{key}.pickle."
                    )
            self.filename_tmpl = filename_tmpl

    #: Where to persist cached values
    filename_tmpl = gettempdir() + '/pld/cache-{module}.{func}-{key}.pickle'

    #: Using numpy_pickle.load
    load = staticmethod(numpy_pickle.load)

    #: Using numpy_pickle.dump
    dump = staticmethod(numpy_pickle.dump)

    @property
    def cache(self):
        return self

    def _filename(self, key):
        hashed_key = hashlib.sha1(str(key).encode('utf-8')).hexdigest()[:8]
        return self.filename_tmpl.format(
            module=self.func.__module__,
            func=self.func.__name__,
            key=hashed_key,
            )

    def __getitem__(self, key):
        filename = self._filename(key)
        if os.path.exists(filename):
            return self.load(filename)
        else:
            raise KeyError(key)

    def __setitem__(self, key, value):
        filename = self._filename(key)
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        return self.dump(value, filename)


class picklediskcache(diskcache):
    """
    Same as diskcache, except that standard pickle is used instead of
    joblib's pickle functionality.
    """
    def load(self, filename):
        with open(filename, 'rb') as f:
            return pickle.load(f)

    def dump(self, value, filename):
        with open(filename, 'wb') as f:
            return pickle.dump(value, f, -1)


def compute_key_attrs(attrs):
    def compute_key(self, *args, **kwargs):
        return tuple(getattr(self, attr) for attr in attrs)
    return compute_key



from contextlib import contextmanager
from copy import deepcopy
from functools import reduce
import operator
import os
import threading
import time
from unittest.mock import patch

import pytest


class MyDummyComponent:
    def __init__(self, arg1, arg2='blargh', subcomponent=None):
        self.arg1 = arg1
        self.arg2 = arg2
        self.subcomponent = subcomponent
        self.initialize_component_arg = None

    def initialize_component(self, config):
        self.initialize_component_arg = config

    def __eq__(self, other):
        return all([
            self.arg1 == other.arg1,
            self.arg2 == other.arg2,
            self.subcomponent == other.subcomponent,
            self.initialize_component_arg == other.initialize_component_arg,
            ])


class BlockingDummy:
    def __init__(self):
        time.sleep(0.1)


class BadDummy:
    def __init__(self):
        from palladium.config import get_config
        self.cfg = get_config().copy()


@contextmanager
def cwd(path):
    before = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(before)


def test_config_class_keyerror():
    from palladium.config import Config
    with pytest.raises(KeyError) as e:
        Config({})['invalid']
    assert "Maybe you forgot to set" in str(e.value)


class TestInitializeConfig:
    @pytest.fixture
    def initialize_config(self):
        from palladium.config import initialize_config
        return initialize_config

    def test_extra(self, config, initialize_config):
        config.clear()
        initialize_config(two='three')
        assert config['two'] == 'three'

    def test_already_initialized(self, config, initialize_config):
        config.clear()
        config.initialized = True
        with pytest.raises(RuntimeError):
            initialize_config(two='three')


class TestGetConfig:
    @pytest.fixture
    def get_config(self):
        from palladium.config import get_config
        return get_config

    @pytest.fixture
    def config1_fname(self, tmpdir):
        path = tmpdir.join('palladium-config.py')
        path.write("""{
            'env': environ['ENV1'],
            'here': here,
            'blocking': {
                '!': 'palladium.tests.test_config.BlockingDummy',
            }
        }""")
        return str(path)

    @pytest.fixture
    def config2_fname(self, tmpdir):
        path = tmpdir.join('config2.py')
        path.write("{'env': environ['ENV2']}")
        return str(path)

    @pytest.fixture
    def config3_fname(self, tmpdir):
        path = tmpdir.join('config3.py')
        path.write("""{
            'bad': {
                '!': 'palladium.tests.test_config.BadDummy'
             }
        }""")
        return str(path)

    def test_extras(self, get_config):
        assert get_config(foo='bar')['foo'] == 'bar'

    def test_default_config(self, get_config, config1_fname, monkeypatch):
        here = os.path.dirname(config1_fname)
        monkeypatch.setitem(os.environ, 'ENV1', 'one')
        with cwd(here):
            config = get_config()
        assert config['here'] == here

    def test_variables(self, get_config, config1_fname, monkeypatch):
        monkeypatch.setitem(os.environ, 'PALLADIUM_CONFIG', config1_fname)
        monkeypatch.setitem(os.environ, 'ENV1', 'one')
        config = get_config()
        assert config['env'] == 'one'
        assert config['here'] == os.path.dirname(config1_fname)

    def test_multiple_files(self, get_config, config1_fname, config2_fname,
                            monkeypatch):
        monkeypatch.setitem(os.environ, 'PALLADIUM_CONFIG',
                            ','.join([config1_fname, config2_fname]))
        monkeypatch.setitem(os.environ, 'ENV1', 'one')
        monkeypatch.setitem(os.environ, 'ENV2', 'two')
        config = get_config()
        assert config['env'] == 'two'
        assert config['here'] == os.path.dirname(config1_fname)

    def test_multithreaded(self, get_config, config1_fname, monkeypatch):
        monkeypatch.setitem(os.environ, 'PALLADIUM_CONFIG', config1_fname)
        monkeypatch.setitem(os.environ, 'ENV1', 'one')

        cfg = {}

        def get_me_config():
            cfg[threading.get_ident()] = get_config().copy()

        threads = [threading.Thread(target=get_me_config) for i in range(2)]
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()

        assert reduce(operator.eq, cfg.values())

    def test_pld_config_key(self, get_config, config1_fname, monkeypatch):
        monkeypatch.setitem(os.environ, 'PALLADIUM_CONFIG', config1_fname)
        monkeypatch.setitem(os.environ, 'ENV1', 'one')
        config = get_config()
        assert config['blocking'].__pld_config_key__ == 'blocking'


class TestProcessConfig:
    @pytest.fixture
    def process_config(self):
        from palladium.config import process_config
        return process_config

    @pytest.fixture
    def config1(self):
        dummy = 'palladium.tests.test_config.MyDummyComponent'
        return {
            'mycomponent': {
                '!': dummy,
                'arg1': 3,
                'arg2': {'no': 'factory'},
                'subcomponent': {
                    '!': dummy,
                    'arg1': {
                        'subsubcomponent': {
                            '!':
                            dummy,
                            'arg1': 'wobwob',
                            'arg2': 9,
                            },
                        },
                    'arg2': 6,
                    },
                },
            'mylistofcomponents': [{
                '!': dummy,
                'arg1': 'wobwob',
                },
                'somethingelse',
                ],
            'mynestedlistofcomponents': [[{
                '!': dummy,
                'arg1': 'feep',
                'arg2': {
                    '__factory__': dummy,  # alternative to '!'
                    'arg1': 6,
                },
            }]],
            'myconstant': 42,

            'mycopiedconstant': {
                '__copy__': 'mycomponent.arg1',
                },

            'mydict': {
                'arg1': 1,
                'mycopiedcomponent': {
                    '__copy__': 'mycomponent',
                    'arg2': None,
                    },
                },

            '__python__': """
C['mynestedlistofcomponents'][0][0]['arg2']['!'] = 'builtins:dict'
C['myotherconstant'] = 13
""",
            }

    @pytest.fixture
    def config2(self):
        return {
            'mydict': {
                '__copy__': 'mydict',
                'arg1': 3,
                'arg2': None,
                },
            'mynewdict': {
                '__copy__': 'mydict',
                'arg2': 2,
                },
            'mysupernewdict': {
                '__copy__': 'mynewdict',
                },
            'mycopiedconstant': {
                '__copy__': 'mycopiedconstant',
                '__default__': 42,
                },
            'mycopywithdefault': {
                '__copy__': 'nonexistant',
                '__default__': 42,
                },
            }

    def test_config1(self, process_config, config1):
        config = process_config(config1)

        assert config['myconstant'] == 42

        mycomponent = config['mycomponent']
        assert isinstance(mycomponent, MyDummyComponent)
        assert mycomponent.arg1 == 3
        assert mycomponent.arg2 == {'no': 'factory'}
        assert mycomponent.initialize_component_arg is config

        subcomponent = mycomponent.subcomponent
        assert isinstance(subcomponent, MyDummyComponent)
        assert subcomponent.arg2 == 6
        assert subcomponent.initialize_component_arg is config

        subsubcomponent = subcomponent.arg1['subsubcomponent']
        assert isinstance(subsubcomponent, MyDummyComponent)
        assert subsubcomponent.arg1 == 'wobwob'
        assert subsubcomponent.arg2 == 9
        assert subsubcomponent.initialize_component_arg is config

        mylistofcomponents = config['mylistofcomponents']
        assert len(mylistofcomponents) == 2
        assert isinstance(mylistofcomponents[0], MyDummyComponent)
        assert mylistofcomponents[0].arg1 == 'wobwob'
        assert mylistofcomponents[1] == 'somethingelse'

        mnl = config['mynestedlistofcomponents']
        assert isinstance(mnl[0][0], MyDummyComponent)
        assert mnl[0][0].arg1 == 'feep'

        assert config['mycopiedconstant'] == 3

        mcc = config['mydict']['mycopiedcomponent']
        assert mcc.arg2 is None
        assert mcc.arg1 == mycomponent.arg1
        assert mcc.subcomponent == mycomponent.subcomponent
        assert mcc.subcomponent is not mycomponent.subcomponent

        assert isinstance(mnl[0][0].arg2, dict)
        assert config['myotherconstant'] == 13

    def test_config1_and_2(self, process_config, config1, config2):
        config = process_config(config1, config2)

        assert config['mydict']['arg1'] == 3

        mycomponent = config['mycomponent']
        mcc = config['mydict']['mycopiedcomponent']
        assert mcc.arg2 is None
        assert mcc.arg1 == mycomponent.arg1
        assert mcc.subcomponent == mycomponent.subcomponent
        assert mcc.subcomponent is not mycomponent.subcomponent

        assert config['mynewdict']['arg1'] == config['mydict']['arg1']
        assert config['mynewdict']['arg2'] == 2
        assert isinstance(
            config['mynewdict']['mycopiedcomponent'], MyDummyComponent)
        assert isinstance(
            config['mysupernewdict']['mycopiedcomponent'], MyDummyComponent)

        assert config['mycopiedconstant'] == 3
        assert config['mycopywithdefault'] == 42

    @pytest.fixture
    def config3(self):
        return {
            'first': 5,
            'second': {
                '__copy__': 'first',
                '__default__': 6,
                },
            }

    def test_copy_source_exists_with_default(self, process_config, config3):
        expected = deepcopy(config3)
        expected['second'] = expected['first']
        got = process_config(config3)
        assert got == expected

    def test_copy_source_exists_no_default(self, process_config, config3):
        expected = deepcopy(config3)
        expected['second'] = expected['first']
        del config3['second']['__default__']
        got = process_config(config3)
        assert got == expected

    def test_copy_source_missing_with_default(self, process_config, config3):
        expected = deepcopy(config3)
        expected['second'] = expected['second']['__default__']
        del expected['first']
        del config3['first']
        got = process_config(config3)
        assert got == expected

    def test_copy_source_missing_no_default(self, process_config, config3):
        del config3['first']
        del config3['second']['__default__']
        with pytest.raises(KeyError):
            process_config(config3)

    def test_initialize_config_logging(self, process_config):
        with patch('palladium.config.dictConfig') as dictConfig:
            process_config({'logging': 'yes, please'})
            dictConfig.assert_called_with('yes, please')
